#!/usr/bin/env python3

import os
import pytest
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_config():

    base = os.path.abspath(os.path.dirname(__file__))
    path = '%s/data/resource_*.json' % base

    with pytest.raises(ValueError):
        _ = ru.Config(path='foo', cfg='bar')

    cfg1 = ru.Config(name=path)

    assert('bar' == cfg1.yale.query('grace.agent_launch_method'))
    assert('bar' == cfg1['yale']['grace']['agent_launch_method'])

    assert(None  is cfg1.query('yale.grace.no_launch_method'))
    with pytest.raises(KeyError):
        _ = cfg1['yale']['grace']['no_launch_method']                # noqa F841

    os.environ['FOO'] = 'GSISSH'

    cfg2 = ru.Config(name=path)
    assert('GSISSH' == cfg2.query('yale.grace.agent_launch_method'))
    assert(None     is cfg2.query('yale.grace.no_launch_method'))

    cfg3 = ru.Config(name=path, expand=False)
    assert('${FOO:bar}' == cfg3.query('yale.grace.agent_launch_method'))

    env  = {'FOO' : 'baz'}
    cfg4 = ru.Config(name=path, env=env)
    assert('baz' == cfg4.query('yale.grace.agent_launch_method'))

    # test `cls._self_default` flag
    cfg5 = ru.Config(from_dict={'foo_0': {'foo_1': {'foo2': 'bar'}}})
    assert('bar' == cfg5.foo_0.foo_1.foo2)
    assert(isinstance(cfg5.foo_0.foo_1, ru.Config))


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_config()


# ------------------------------------------------------------------------------






import threading as mt

from ..logger  import Logger
from ..profile import Profiler
from ..config  import Config
from ..json_io import read_json, write_json


# ------------------------------------------------------------------------------
#
class Bridge(object):
    '''

    A bridge can be configured to have a finite lifetime: when no messages are
    received in `timeout` seconds, the bridge process will terminate.
    '''

    # --------------------------------------------------------------------------
    #
    @staticmethod
    def get_config(name, pwd=None):

        if not pwd:
            pwd = '.'

        cfg = Config(cfg=read_json('%s/%s.cfg' % (pwd, name)))

        return cfg


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        self._cfg     = cfg
        self._channel = self._cfg.channel
        self._uid     = self._cfg.uid
        self._pwd     = self._cfg.path
        self._log     = Logger(name=self._uid, ns='radical.utils',
                               level=self._cfg.log_lvl, path=self._pwd)
        self._prof    = Profiler(name=self._uid, path=self._pwd)

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()
        else:
            self._prof.disable()

        self._prof.prof('init', uid=self._uid, msg=self._pwd)
        self._log.debug('bridge %s init', self._uid)

        self._bridge_initialize()

        write_json('%s/%s.cfg' % (self._pwd, self._cfg.uid),
                   {'uid'        : self._cfg.uid,
                    self.type_in : str(self.addr_in),
                    self.type_out: str(self.addr_out)})


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel

    # protocol independent addr query
    @property
    def type_in(self):
        raise NotImplementedError()

    @property
    def type_out(self):
        raise NotImplementedError()

    @property
    def addr_in(self):
        raise NotImplementedError()

    @property
    def addr_out(self):
        raise NotImplementedError()

    def _bridge_initialize(self):
        raise NotImplementedError()

    def _bridge_work(self):
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    #
    def start(self):

        # the bridge runs in a thread.  It is the bridge's owner process'
        # responsibility to ensure the thread is seeing suffient time to perform
        # as needed.  Given Python's thread performance (or lack thereof), this
        # basically means that the user of this class should create a separate
        # process instance to host the bridge thread.
        self._term          = mt.Event()
        self._bridge_thread = mt.Thread(target=self._bridge_work)
        self._bridge_thread.daemon = True
        self._bridge_thread.start()

        self._log.info('started bridge %s', self._uid)


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def create(cfg):

        # NOTE: I'd rather have this as class data than as stack data, but
        #       python stumbles over circular imports at that point :/
        #       Another option though is to discover and dynamically load
        #       components.
        from .pubsub import PubSub
        from .queue  import Queue

        _btypemap = {'pubsub' : PubSub,
                     'queue'  : Queue}

        kind = cfg['kind']

        if kind not in _btypemap:
            raise ValueError('unknown bridge type (%s)' % kind)

        btype  = _btypemap[kind]
        bridge = btype(cfg)

        return bridge


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._term.set()
      # self._bridge_thread.join(timeout=timeout)
        self._prof.prof('term', uid=self._uid)

      # if timeout is not None:
      #     return not self._bridge_thread.is_alive()


    # --------------------------------------------------------------------------
    #
    @property
    def alive(self):
        return self._bridge_thread.is_alive()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pylint: disable=protected-access

import pytest
import radical.utils as ru


# ------------------------------------------------------------------------------
#
class _TestDescr(ru.Description):

    _schema = {'exe'  : str,
               'procs': int,
               'env'  : {str: int},
               'pre'  : [str],
               'post' : (str, ),
               'meta' : None,
               'items': [int],
               '_foo' : str,
               '_data': str,
               'unset': int,
               'i245' : [str],
    }

    def _verify(self):

        if 'exe' not in self:
            raise ValueError('exe not defined')


# ------------------------------------------------------------------------------
#
def test_description():

    from_dict = {'procs': '3',
                 'env'  : {3: '4'},
                 'pre'  : True,
                 'post' : False,
                 'meta' : 3.4,
                 'items': [3.4, '3'],
                 '_foo' : ['bar'],
                 '_data': 4,
                 'i245' : [],
    }

    td = _TestDescr()
    td = _TestDescr(from_dict=from_dict)

    with pytest.raises(ValueError):
        td.verify()

    td.exe = '/bin/date'
    assert(td.verify())

    assert(isinstance(td.exe,   str))
    assert(isinstance(td.procs, int))
    assert(isinstance(td.env,   dict))
    assert(isinstance(td.pre,   list))
    assert(isinstance(td.post,  tuple)), td.post
  # assert(isinstance(td._foo,  str)), td._foo                            # noqa

    assert(isinstance(td['_data'], str))
    assert(isinstance(td['items'], list))

    assert(td.meta)

    assert(td.exe      == '/bin/date')
    assert(td.procs    == 3)
    assert(td.env      == {'3': 4})
    assert(td.pre      == ['True'])
    assert(td.post     == ('False',))
  # assert(td._foo     == "['bar']")                                      # noqa

    assert(td['_data'] == '4')
    assert(td['items'] == [3, 3])

    assert(td.meta     == 3.4)

    import copy
    c = copy.deepcopy(td)
    assert(c.as_dict() == td.as_dict())

    td.i245.append('foo')
    td = _TestDescr()
    assert(not td.i245), td.i245


# ------------------------------------------------------------------------------
#
def test_inheritance():

    class _Test(ru.Description):

        _schema   = {'i245': [str]}
        _defaults = {'i245': []   }

        def __init__(self, from_dict=None):
            ru.Description.__init__(self, from_dict=from_dict)

    # --------------------------------------------------------------------------
    td = _Test()
    assert(not td.i245)

    td.i245.append('foo')
    assert(td.i245 == ['foo'])

    td = _Test()
    assert (not td.i245), td.i245


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_description()
    test_inheritance()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import os
os.environ['RADICAL_BASE'] = '/tmp'

import glob
import shutil
import time
from uuid import UUID

import radical.utils as ru

from unittest import TestCase

TEST_CASES_PATH = 'tests/test_cases/ids.*.json'


class IdsTestClass(TestCase):

    # --------------------------------------------------------------------------
    #
    @classmethod
    def setUpClass(cls):
        cls._base_dir = ru.get_radical_base('utils')
        cls._pid_str = '%06d' % os.getpid()
        cls._user = None
        try:
            import getpass
            cls._user = getpass.getuser()
        except:
            cls._user = 'nobody'

        cls._test_cases = []
        for f in glob.glob(TEST_CASES_PATH):
            cls._test_cases.extend(ru.read_json(f))

    # --------------------------------------------------------------------------
    #
    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, '_base_dir') and os.path.isdir(cls._base_dir):
            try:
                shutil.rmtree(cls._base_dir)
            except OSError as e:
                print('[ERROR] %s - %s' % (e.filename, e.strerror))

    # --------------------------------------------------------------------------
    #
    def setUp(self):
        pass

    # --------------------------------------------------------------------------
    #
    def tearDown(self):
        ru.reset_id_counters(prefix=None, reset_all_others=True)

    # --------------------------------------------------------------------------
    #
    def test_generate_id(self):

        for test_case in self._test_cases:

            # check possible exceptions firstly
            if 'exception' in test_case['result']:
                if test_case['result']['exception'] == 'TypeError':
                    _exc = TypeError
                elif test_case['result']['exception'] == 'ValueError':
                    _exc = ValueError
                else:
                    _exc = Exception

                self.assertRaises(_exc, ru.generate_id, **test_case['input'])
                continue

            # generate ID
            id_str = ru.generate_id(**test_case['input'])

            # if it is assumed that associated file will be created
            file_path = None
            if 'file_name_template' in test_case['result']:
                file_path_parts = [self._base_dir]
                if test_case['input'].get('ns'):
                    file_path_parts.append(test_case['input']['ns'])
                file_path_parts.append(
                    # here is a dict with all possible parameters that might be
                    # in the template name of the corresponding file (counter)
                    test_case['result']['file_name_template'] % {
                        'user': self._user,
                        'day_counter': int(time.time() / (60 * 60 * 24))})
                file_path = os.path.join(*file_path_parts)

            # start to go through test cases
            if not test_case['input'].get('mode') or \
                    test_case['input']['mode'] == ru.ID_SIMPLE:
                self.assertEqual(id_str, test_case['result']['id'])

            elif test_case['input']['mode'] == ru.ID_UNIQUE:
                self.assertTrue(id_str.endswith(test_case['result']['id_end']))
                self.assertIn(self._pid_str, id_str)

            elif test_case['input']['mode'] == ru.ID_UUID:
                uuid_str = id_str.split('%s.' % test_case['input']['prefix'])[1]
                self.assertIsInstance(UUID(uuid_str), UUID)

            elif test_case['input']['mode'] == ru.ID_PRIVATE:
                self.assertIn(self._user, id_str)
                # check that corresponding file was created
                self.assertTrue(os.path.isfile(file_path))

            elif test_case['input']['mode'] == ru.ID_CUSTOM:

                if '%(day_counter)' in test_case['input']['prefix']:
                    # check that created file contains the correct value
                    with open(file_path) as fd:
                        # file with counter contains the number of the next ID
                        # (ID was generated and the counter was increased)
                        counter = '%d' % (int(fd.readline()) - 1)
                    self.assertIn(counter, id_str)

                elif '%(item_counter)' in test_case['input']['prefix']:
                    # check that corresponding file was created
                    self.assertTrue(os.path.isfile(file_path))

    def test_generate_id_2nd_run(self):
        # check that counters got reset
        self.test_generate_id()

# ------------------------------------------------------------------------------




# ------------------------------------------------------------------------------
#
# We provide a base class for all kinds of description objects in the RCT
# stack: job descriptions, task descriptions, pilot descriptions, workload
# descriptions etc.  That base class provides:
#
#   - dict like API
#   - schema based type definitions
#   - optional runtime type checking
#
# The Description base class provides a property API, similar to the `ru.Config`
# class.
#

from .munch import Munch


# ------------------------------------------------------------------------------
#
class Description(Munch):
    '''
    This is an abstract base class for RCT description types.  Any inheriting
    class MUST provide a `self._schema` class member (not class instance member)
    which is used to verify the description's data validity.  Validation can be
    performed on request (`d.verify()`), or when setting description properties.
    The default is to verify on explicit calls only.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None, verify_setter=False):

        super(Description, self).__init__(from_dict=from_dict)

        if verify_setter:
            raise NotImplementedError('setter verification not yet implemented')

        # TODO: setter verification should be done by attempting to cast the
        #       value to the target type and raising on failure.  Non-trivial
        #       types (lists, dicts) can use `as_list` and friends, or
        #       `isinstance` if that is not available
        #
        # TODO: setter verification should verify that the property is allowed


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import sys
import time
import pytest


import multiprocessing as mp
import radical.utils   as ru


fname = '/tmp/ru.test.lockfile.%s' % os.getpid()


def _get_lock(_fname, delay, timeout, duration):
    lock = ru.Lockfile(_fname, delete=True)
    try:
        time.sleep(delay)
        with lock(timeout=timeout):
            time.sleep(duration)
            sys.exit(1)
    except TimeoutError:
        sys.exit(2)
    except RuntimeError:
        sys.exit(3)


# ------------------------------------------------------------------------------
#
def test_lockfile():

    try:
        with ru.Lockfile(fname, owner='foo') as fd0:
            assert(fd0)
            assert(fd0.get_owner() == 'foo'), fd0.get_owner()
            fd0.write('test 0\n')

        with ru.Lockfile(fname) as fd1:
            assert(fd1)
            fd1.lseek(0, os.SEEK_SET)
            fd1.write('test 1\n')

        with ru.Lockfile(fname) as fd2:
            assert(fd2)
            fd2.lseek(0, os.SEEK_END)
            fd2.write('test 2\n')

            with pytest.raises(RuntimeError):
                with ru.Lockfile(fname) as fd3:
                    fd3.lseek(0, os.SEEK_END)
                    fd3.write('test 3\n')

        with open(fname, 'r') as fin:
            data = fin.read()

            assert(data == 'test 1\ntest 2\n')

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_lockfile_timer():

    try:

        with ru.Lockfile(fname) as fd0:
            assert(fd0)
            fd0.write('test 0\n')

            fd1   = None
            start = time.time()
            with pytest.raises(TimeoutError):
                fd1 = ru.Lockfile(fname)
                fd1.acquire(timeout=0.1)

            stop = time.time()
            assert(stop - start > 0.1)
            assert(stop - start < 0.2)

            if fd1:
                try:
                    fd1.release()
                except:
                    pass

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_lockfile_ok():

    try:

        p1 = mp.Process(target=_get_lock, args=[fname, 0.0, 0.0, 0.2])
        p2 = mp.Process(target=_get_lock, args=[fname, 0.1, 0.3, 0.0])

        p1.start()
        p2.start()

        p1.join()
        p2.join()

        assert(p1.exitcode == 1), p1.exitcode
        assert(p2.exitcode == 1), p2.exitcode

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_lockfile_nok():

    try:

        p1 = mp.Process(target=_get_lock, args=[fname, 0.0, 0.0, 0.3])
        p2 = mp.Process(target=_get_lock, args=[fname, 0.1, 0.1, 0.1])

        p1.start()
        p2.start()

        p1.join()
        p2.join()

        assert(p1.exitcode == 1), p1.exitcode
        assert(p2.exitcode == 2), p2.exitcode

    finally:
        try   : os.unlink(fname)
        except: pass


    try:

        p1 = mp.Process(target=_get_lock, args=[fname, 0.0, 0.0, 0.5])
        p2 = mp.Process(target=_get_lock, args=[fname, 0.1, 0.0, 0.0])

        p1.start()
        p2.start()

        p1.join()
        p2.join()

        assert(p1.exitcode == 1), p1.exitcode
        assert(p2.exitcode == 3), p2.exitcode

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_lockfile_delete():

    try:
        p1 = mp.Process(target=_get_lock, args=[fname, 0.0, 0.0, 0.5])
        p1.start()
        time.sleep(0.1)
        assert(os.path.isfile(fname))

        p1.join()
        assert(not os.path.isfile(fname))
        assert(p1.exitcode == 1)

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_lockfile()
    test_lockfile_timer()
    test_lockfile_ok()
    test_lockfile_nok()
    test_lockfile_delete()

# ------------------------------------------------------------------------------




# pylint: disable=unused-import,eval-used

__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import re
import json

from .misc import as_string


# ------------------------------------------------------------------------------
#
def read_json(fname, filter_comments=True):
    '''
    Comments line in the form of

        # some json data or text

    are stripped from json before parsing:

        import pprint
        pprint.pprint(read_json("my_file.json"))
    '''

    with open(fname) as f:

        try:
            return parse_json(f.read(), filter_comments)

        except ValueError as e:
            raise ValueError('error parsing %s: %s' % (fname, e)) from e


# ------------------------------------------------------------------------------
#
def read_json_str(filename, filter_comments=True):
    '''
    same as read_json, but converts unicode strings and byte arrays to ASCII
    strings.
    '''

    return as_string(read_json(filename, filter_comments))


# ------------------------------------------------------------------------------
#
def write_json(data, fname):
    '''
    thin wrapper around python's json write, for consistency of interface

    '''

    if isinstance(fname, dict) and isinstance(data, str):
        # arguments were switched: accept anyway
        tmp   = data
        data  = fname
        fname = tmp

    with open(fname, 'w') as f:
        json.dump(data, f, sort_keys=True, indent=4, ensure_ascii=False)
        f.write('\n')
        f.flush()


# ------------------------------------------------------------------------------
#
def parse_json(json_str, filter_comments=True):
    '''
    Comment lines in the form of

        # some json data or text

    are stripped from json before parsing
    '''

    if not filter_comments:
        return json.loads(json_str)

    else:
        content = ''
        for line in json_str.split('\n'):
            content += re.sub(r'^\s*#.*$', '', line)
            content += '\n'

        return json.loads(content)


# ------------------------------------------------------------------------------
#
def parse_json_str(json_str):
    '''
    same as parse_json, but converts unicode strings to simple strings
    '''

    return as_string(parse_json(json_str))


# ------------------------------------------------------------------------------
#
def metric_expand(data):
    '''
    iterate through the given dictionary, and when encountering a key string of
    the form `ru.XYZ` or `rp.ABC`, expand them to their actually defined values.
    This the following dict:

        {
            "ru.EVENT" : "foo"
        }

    becomes:


        {
            2 : "foo"
        }

    '''

    try   : import radical.pilot as rp                              # noqa: F401
    except: pass
    try   : import radical.saga  as rs                              # noqa: F401
    except: pass
    try   : import radical.utils as ru                              # noqa: F401
    except: pass

    if isinstance(data, str):

        if data.count('.') == 1:
            elems = data.split('.')
            if len(elems[0]) == 2 and elems[0][0] == 'r':
                try:
                    data = eval(data)
                finally:

                    pass
        return data

    elif isinstance(data, list):
        return [metric_expand(elem) for elem in data]

    elif isinstance(data, dict):
        return {metric_expand(k) : metric_expand(v) for k,v in data.items()}

    else:
        return data


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import pytest
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_munch():
    # note that test_description.py also tests large parts of the Munch class in
    # a non-hierarchical setup.  This test is focused on inheritance: note that
    # all classes inherit from ru.Config which is also Munch derivative, but
    # provides simpler initialization which we use in these tests.

    # --------------------------------------------------------------------------
    # empty class
    class Foo(ru.Config):
        _schema = {'one': int,
                   'two': {str: int}}

    # --------------------------------------------------------------------------
    # plain schema
    class Bar_1(ru.Config):
        _schema = {'one': int,
                   'two': {str: int}}

    # --------------------------------------------------------------------------
    # class whose schema is composed
    class Buz_1(ru.Config):
        _schema = {'foo': int,
                   'bar': Bar_1}


    cfg = {'foo'   : 2,
           'bar'   : {'one': 3,
                      'two': {'buz': 2}}}
    r = Buz_1(cfg=cfg)
    r.verify()
    assert(type(r.bar).__name__ == 'Bar_1')
    assert(r.as_dict() == cfg), [cfg, r.as_dict()]


    # confirm that checks are active
    cfg = {'foo'   : 'foo',
           'bar'   : {'one': 3,
                      'two': {'buz': 2}}}
    r = Buz_1(cfg=cfg)
    with pytest.raises(TypeError):
        # will not be able to convert attribute "foo" (cast=True)
        r.verify()


    # confirm that composition is checked
    cfg = {'foo'   : 2,
           'bar'   : {'buz':  2}}
    r = Buz_1(cfg=cfg)
    with pytest.raises(TypeError):
        # will not be able to convert attribute "bar" (cast=True)
        r.verify()


    # confirm that typed composition works
    cfg = {'foo'   : 2,
           'bar'   : Bar_1(cfg={'buz':  2})}
    r = Buz_1(cfg=cfg)
    with pytest.raises(TypeError):
        # will not be able to convert input data for `Bar_1` (cast=True)
        r.verify()


    # confirm that typed composition is checked
    cfg = {'foo'   : 2,
           'bar'   : Bar_1(cfg={'buz':  'buz'})}
    r = Buz_1(cfg=cfg)
    with pytest.raises(TypeError):
        # will not be able to convert input data for `Bar_1` (cast=True)
        r.verify()


    # confirm that composition type is checked
    cfg = {'foo'   : 2,
           'bar'   : Foo()}
    r = Buz_1(cfg=cfg)
    # no TypeError, since value of attribute "bar" will be converted
    # classes `Foo` and `Bar_1` have the same schemas
    r.verify()


    # confirm that nested types are checked
    cfg = {'foo'   : 2,
           'bar'   : {'one': 3,
                      'two': {'buz': 'bar'}}}
    r = Buz_1(cfg=cfg)
    with pytest.raises(TypeError):
        # will not be able to convert input data for `Bar_1` (attribute `two`)
        r.verify()

    # --------------------------------------------------------------------------
    # check if an inherited schema can be extended
    class Buz_2(Buz_1):
        # extend schema: add a type
        Buz_1._schema['biz'] = {str: int}

    cfg = {'foo'   : 2,
           'biz'   : {'buz': 0},
           'bar'   : {'one': 3,
                      'two': {'buz': 1}}}
    r = Buz_2(cfg=cfg)
    r.verify()

    # confirm that new type is checked
    cfg = {'foo'   : 2,
           'biz'   : {'buz': 'buz'},
           'bar'   : {'one': 3,
                      'two': {'buz': 1}}}
    r = Buz_2(cfg=cfg)
    with pytest.raises(TypeError):
        r.verify()

    # confirm that old type is checked
    cfg = {'foo'   : 2,
           'biz'   : {'buz': 0},
           'bar'   : {'one': 3,
                      'two': {'buz': 'buz'}}}
    r = Buz_2(cfg=cfg)
    with pytest.raises(TypeError):
        r.verify()


# ------------------------------------------------------------------------------
#
def test_munch_defaults():

    # --------------------------------------------------------------------------
    # plain schema
    class Foo(ru.Munch):
        _schema   = {'one'  : int,
                     'two'  : {str: int},
                     'three': [str],
                     'four' : float,
                     'five' : bool}
        _defaults = {'two'  : {'two-default': 0},
                     'three': [3, 3, 3],
                     'four' : None,
                     'five' : True}
    # --------------------------------------------------------------------------


    f = Foo()
    assert (f.one   is None)
    assert (f.two   == {'two-default': 0})
    assert (f.three == [3, 3, 3])
    assert (f.four  is None)
    assert (f.five  is True)


# ------------------------------------------------------------------------------
#
def test_munch_update():

    # --------------------------------------------------------------------------
    # plain schema
    class Bar_1(ru.Munch):
        _schema   = {'one'  : int,
                     'two'  : {str: int},
                     'three': [str]}
        _defaults = {'two'  : {'two-default': 0},
                     'three': [3, 3, 3]}

    # --------------------------------------------------------------------------
    # class whose schema is composed
    class Buz_1(ru.Munch):
        _schema = {'foo': int,
                   'bar': Bar_1}

    # --------------------------------------------------------------------------
    # class whose schema is composed and has default values
    class Foo_1(ru.Munch):
        _schema   = {'buz': Buz_1}
        _defaults = {'buz': {'foo': 0,
                             'bar': {'one': 1,
                                     'two': {'sub-two': 2}}}}

    # --------------------------------------------------------------------------
    f = Foo_1()
    assert (f.buz.foo == 0)
    assert (f.buz.bar.two['sub-two'] == 2)
    assert (f.buz.bar.three == [3, 3, 3])

    f = Foo_1({'buz': {'foo': 3,
                       'bar': {'one': 11}}})
    assert (f.buz.bar.one == 11)
    assert (f.buz.bar.two['sub-two'] == 2)

    f = Foo_1({'buz': {'foo': 3,
                       'bar': {'one': 0,
                               'two': {2: 22}}}})
    f.verify()  # method `verify` convert int into str according to the schema
    assert (f.buz.bar.two['2'] == 22)

    f = Foo_1()
    b = Bar_1()
    f.update({'buz': {'bar': b}})
    # default values from Bar_1 class
    assert (f.buz.bar.three == [3, 3, 3])
    assert (f.buz.bar['two'] == {'two-default': 0})
    assert (f.buz.bar.get('one') is None)

    # test `cls._self_default` flag (for method `update`)

    # --------------------------------------------------------------------------
    # cls._self_default is False
    class Bar_2a(ru.Munch):
        _self_default = False

    # --------------------------------------------------------------------------
    # cls._self_default is True
    class Bar_2b(ru.Munch):
        _self_default = True

    # --------------------------------------------------------------------------
    bar = Bar_2a(from_dict={'foo_0': {'foo_1': {'foo2': 'bar'}}})
    assert (not isinstance(bar.foo_0, Bar_2a))
    assert (isinstance(bar.foo_0.foo_1, ru.Munch))

    bar = Bar_2b(from_dict={'foo_0': {'foo_1': {'foo2': 'bar'}}})
    assert (isinstance(bar.foo_0, Bar_2b))
    assert (issubclass(type(bar.foo_0.foo_1), ru.Munch))


# ------------------------------------------------------------------------------
#
def test_munch_inherit():

    # --------------------------------------------------------------------------
    # plain schema
    class Foo(ru.Munch):
        _schema   = {'one'  : int,
                     'two'  : {str: int},
                     'three': [str]}
        _defaults = {'two'  : {'two-default': 0},
                     'three': [3, 3, 3]}

    # --------------------------------------------------------------------------
    # class whose schema is composed
    class Bar(Foo):
        _schema_extend   = {'four': int,
                            'five': float,
                            'six' : Foo}
        _defaults_extend = {'four': 42,
                            'six' : {'one'  : 1,
                                     'three': [33, 33]}}

    # --------------------------------------------------------------------------
    b = Bar()

    assert (b.one                is None)
    assert (b.two['two-default'] == 0)
    assert (b.three              == [3, 3, 3])
    assert (b.four               == 42), b.four
    assert (b.five               is None)
    assert (b.six.one            == 1)
    assert (b.six.three[1]       == 33)


# ------------------------------------------------------------------------------
#
def test_demunch():

    # --------------------------------------------------------------------------
    #
    class Bar_1(ru.Munch):
        _schema = {'bars': list}

    # --------------------------------------------------------------------------
    #
    class Buz_1(ru.Munch):
        _schema = {'foo': int,
                   'bar': Bar_1,
                   'buz': tuple}

    # --------------------------------------------------------------------------

    b = Bar_1({'bars': ['bar_1', Bar_1({'bars': [1, 2, 3]})]})
    b_dict = b.as_dict()
    assert isinstance(b_dict,            dict)
    assert not isinstance(b_dict,        Bar_1)
    assert isinstance(b_dict['bars'],    list)
    assert isinstance(b_dict['bars'][1], dict)
    assert b_dict['bars'][0] == 'bar_1'
    assert len(b_dict['bars'][1]['bars']) == 3

    b = Buz_1({'foo': 3, 'bar': Bar_1({'bars': [3, 2, 1]}), 'buz': tuple([5])})
    b_dict = b.as_dict()
    assert isinstance(b_dict,                dict)
    assert isinstance(b_dict['bar'],         dict)
    assert isinstance(b_dict['bar']['bars'], list)
    assert isinstance(b_dict['buz'],         tuple)
    assert not isinstance(b_dict['bar'],     Bar_1)
    assert b_dict['foo'] == 3
    assert b_dict['buz'][0] == 5


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_munch()
    test_munch_defaults()
    test_munch_update()
    test_munch_inherit()
    test_demunch()

# ------------------------------------------------------------------------------





# pylint: disable=raise-missing-from

# ------------------------------------------------------------------------------
#
# We provide a base class for all kinds of dict-like data objects in the RCT
# stack: resource configs, agent configs, job descriptions, task descriptions,
# pilot descriptions, workload descriptions etc.  That class provides:
#
#   - dict like API
#   - public dict keys exposed as attributes
#   - schema based type definitions
#   - optional runtime type checking
#

import copy

from .misc       import as_list, as_tuple
from .misc       import is_string
from .misc       import expand_env as ru_expand_env
from .dict_mixin import DictMixin, dict_merge
from .json_io    import write_json


# ------------------------------------------------------------------------------
#
class Munch(DictMixin):

    _self_default = False

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):
        '''
        create a munchified dictionary (tree) from `from_dict`.

        from_dict: data to be used for initialization

        NOTE: the names listed below are valid keys when used via the
              dictionary API, and can also be *set* via the property API, but
              they cannot be *queried* via the property API as their names
              conflict with the class method names:

                  as_dict
                  clear
                  get
                  has_key
                  items
                  iterkeys
                  itervalues
                  keys
                  popitem
                  setdefault
                  update
                  values
                  verify

              Names with a leading underscore are not supported.
        '''

        if not hasattr(self, '_check'):
            # check type checking on set
            self._check = False

        if not hasattr(self, '_cast'):
            # attempt to cast on type mismatch
            self._cast = True

        if not hasattr(self, '_schema'):
            self._schema = dict()
          # raise RuntimeError('class %s has no schema defined' % self.__name__)

        if hasattr(self, '_schema_extend'):
            self._schema.update(self._schema_extend)

        self._data = dict()

        if hasattr(self, '_defaults'):
            self.update(copy.deepcopy(self._defaults))

        if hasattr(self, '_defaults_extend'):
            self.update(copy.deepcopy(self._defaults_extend))

        self.update(from_dict)


    # --------------------------------------------------------------------------
    #
    def update(self, other):
        '''
        we overload `DictMixin.update()`: the call is used to ensure that
        sub-dicts are instantiated as their respective Munch-inheriting class
        types if so specified by the respective schema.  So, if the schema
        contains:

          {
            ...
            'foo': BarMunch
            ...
          }

        where `BarMunch` is a valid type in the scope of the schema definition
        which inherits from `ru.Munch`, then `update()` will ensure that the
        value for key `foo` is indeed of type `BarMunch`.  An error will be
        raised if (a) `BarMunch` does not have a single parameter constructor
        like `Munch`, or (b) the `data` value for `foo` cannot be used as
        `from_dict` parameter to the `BarMunch` constructor.
        '''

        if not other:
            return

        for k, v in other.items():
            if isinstance(v, dict):
                t = self._schema.get(k)
                if not t:
                    if self._self_default: t = type(self)
                    else                 : t = Munch
                if isinstance(t, type) and \
                        issubclass(t, Munch) and not issubclass(type(v), Munch):
                    # cast to expected Munch type
                    self._data.setdefault(k, t()).update(v)
                    continue
            self[k] = v


    # --------------------------------------------------------------------------
    #
    # base functionality for the DictMixin
    #
    def __getitem__(self, k):
        return self._data[k]

    def __setitem__(self, k, v):
        self._data[k] = self._verify_setter(k, v)

    def __delitem__(self, k):
        del self._data[k]

    def keys(self):
        return self._data.keys()

    def __deepcopy__(self, memo):
        # should return a new instance of the same type, not an original Munch,
        # otherwise if an instance of Munch-based has an attribute of another
        # Munch-based type then `verify` method will raise TypeError exception
        return type(self)(from_dict=copy.deepcopy(self._data))


    # --------------------------------------------------------------------------
    #
    # base functionality for attribute access
    # TODO: Make optional
    #
    def __getattr__(self, k):

        if k.startswith('_'):
            return object.__getattribute__(self, k)

        data   = object.__getattribute__(self, '_data')
        schema = object.__getattribute__(self, '_schema')

        if   not  schema: return data.get(k)
        elif k in schema: return data.get(k)
        else            : return data[k]


    def __setattr__(self, k, v):

        if k.startswith('_'):
            return object.__setattr__(self, k, v)

        self._data[k] = self._verify_setter(k, v)


    def __delattr__(self, k):

        if k.startswith('_'):
            return object.__delattr__(self, k)

        del(self._data[k])


    def __dir__(self):

        return self._data.keys()


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return str(self._data)


    # --------------------------------------------------------------------------
    #
    def __repr__(self):

        return str(self)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _demunch_value(cls, v):
        return v.as_dict() if isinstance(v, Munch) else cls.demunch(v)

    @classmethod
    def demunch(cls, src):
        '''
        iterate given object and apply `Munch.as_dict()` to all munch type
        values, and return the result (effectively a shallow copy).
        '''
        if isinstance(src, dict):
            tgt = {k: cls._demunch_value(v) for k, v in src.items()}
        elif isinstance(src, list):
            tgt = [cls._demunch_value(x) for x in src]
        elif isinstance(src, tuple):
            tgt = tuple([cls._demunch_value(x) for x in src])
        else:
            tgt = src
        return tgt

    def as_dict(self):
        return self.demunch(self._data)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _verify_int(cls, k, v, t, cast):
        if v is None or isinstance(v, int): return v
        if cast:
            try   : return int(v)
            except: raise TypeError('%s: expected int type for %s (%s)'
                                   % (cls.__name__, k, type(v)))
        else:
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_str(cls, k, v, t, cast):
        if v is None or isinstance(v, str): return v
        if cast:
            try   : return str(v)
            except: raise TypeError('%s: expected str type for %s (%s)'
                                   % (cls.__name__, k, type(v)))
        else:
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_float(cls, k, v, t, cast):
        if v is None or isinstance(v, float): return v
        if cast:
            try   : return float(v)
            except: raise TypeError('%s: expected float type for %s (%s)'
                                   % (cls.__name__, k, type(v)))
        else:
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_bool(cls, k, v, t, cast):
        if v is None or isinstance(v, bool): return v
        if cast:
            if str(v).lower() in ['true', 'yes', '1']: return True
            if str(v).lower() in ['false', 'no', '0']: return False
            raise TypeError('%s: expected bool type for %s (%s)'
                           % (cls.__name__, k, type(v)))
        else:
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_tuple(cls, k, v, t, cast):
        if v is None: return
        if cast:
            v = as_tuple(v)
            return tuple([cls._verify_kvt(k + ' tuple element', _v, t[0], cast)
                          for _v in v])
        else:
            if isinstance(v, tuple):
                return v
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_list(cls, k, v, t, cast):
        if v is None: return
        if cast:
            v = as_list(v)
            return [cls._verify_kvt(k + ' list element', _v, t[0], cast)
                    for _v in v]
        else:
            if isinstance(v, list):
                return v
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_dict(cls, k, v, t, cast):
        if v is None: return
        if cast:
            t_k = list(t.keys())[0]
            t_v = list(t.values())[0]
            return {cls._verify_kvt(_k, _k, t_k, cast) :
                    cls._verify_kvt(_k, _v, t_v, cast)
                        for _k, _v in v.items()}
        else:
            if isinstance(v, dict):
                return v
            raise TypeError('attribute type error for %s: expected %s, got %s'
                           % (k, t, type(v)))

    @classmethod
    def _verify_munch(cls, k, v, t, cast):
        if v is None: return
        if cast:
            if issubclass(type(v), t): return v.verify()
            if isinstance(v, dict)   : return t(from_dict=v).verify()
            raise TypeError('attribute type error for %s: expected %s, got %s'
                            % (k, t, type(v)))
        else:
            if issubclass(type(v), t):
                return v
            raise TypeError('attribute type error for %s: expected %s, got %s'
                            % (k, t, type(v)))

    _verifiers = {
            int  : _verify_int.__func__,
            str  : _verify_str.__func__,
            float: _verify_float.__func__,
            bool : _verify_bool.__func__,
    }

    _verifier_keys = list(_verifiers.keys())


    @classmethod
    def _verify_kvt(cls, k, v, t, cast):
        if t is None              : return v
        if t in cls._verifier_keys: return cls._verifiers[t](cls, k, v, t, cast)
        if isinstance(t, tuple)   : return cls._verify_tuple(k, v, t, cast)
        if isinstance(t, list)    : return cls._verify_list(k, v, t, cast)
        if isinstance(t, dict)    : return cls._verify_dict(k, v, t, cast)
        if issubclass(t, Munch)   : return cls._verify_munch(k, v, t, cast)
        if cast: return v
        raise TypeError('%s: no verifier defined for type %s'
                       % (cls.__name__, t))

    def verify(self):
        if self._schema:
            for k, v in self._data.items():
                if k.startswith('__'):
                    continue
                if k not in self._schema:
                    raise TypeError('%s: key %s not in schema'
                                   % (type(self).__name__, k))
                self._data[k] = self._verify_kvt(k, v, self._schema[k],
                                                 self._cast)
        self._verify()
        return self


    # --------------------------------------------------------------------------
    #
    def _verify_setter(self, k, v):

        if not self._check:
            # no type checking on set
            return v

        if self._schema:

            if k not in self._schema:
                raise TypeError('%s: key %s not in schema'
                               % (type(self).__name__, k))
            return self._verify_kvt(k, v, self._schema[k], self._cast)


    # --------------------------------------------------------------------------
    #
    def _verify(self):
        '''
        Can be overloaded
        '''
        pass



    # --------------------------------------------------------------------------
    #
    def merge(self, src, expand=True, env=None, policy='overwrite', log=None):
        '''
        merge the given munch into the existing config settings, overwriting
        any values which already existed
        '''

        if expand:
            # NOTE: expansion is done on reference, not copy
            ru_expand_env(src, env=env)

        dict_merge(self, src, policy=policy, log=log)


    # --------------------------------------------------------------------------
    #
    def write(self, fname):

        write_json(self.as_dict(), fname)


    # --------------------------------------------------------------------------
    #
    def query(self, key, default=None):
        '''
        For a query like

            munch.query('some.path.to.key', 'foo')

        this method behaves like:

            munch['some']['path']['to'].get('key', default='foo')
        '''

        if is_string(key): elems = key.split('.')
        else             : elems = key

        if not elems:
            raise ValueError('empty key on query')

        pos  = self
        path = list()
        for elem in elems:

            if not isinstance(pos, dict):
                raise KeyError('no such key [%s]' % '.'.join(path))

            if elem in pos: pos = pos[elem]
            else          : pos = None

            path.append(elem)

        if pos is None:
            pos = default

        return pos


# ------------------------------------------------------------------------------

demunch = Munch.demunch

# ------------------------------------------------------------------------------





import re
import os

_pat = r'^\s*(?P<detail>(?P<short>[^-]*)(?P<base>-[^-@]+?)?(-[^@]+?)?(?P<branch>@.+?)?)\s*$'


# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version_short :  1.2.3                       - used for installation
#   - version_detail:  1.2.3-v1.1-9-g0684b06@devel - used at runtime (ru.version)
#   - both are derived from the last git tag and branch information
#   - VERSION files are created on install, with the version_detail
#
def get_version(paths=None):
    """
    paths:
        a VERSION file containing the detailed version is checked for in every
        directory listed in paths.   When we find a VERSION file, we also look
        for an SDIST file, and return the found name and location as absolute
        path to the sdist.
    """

    if not paths:
        # by default, get version for myself
        pwd   = os.path.dirname(__file__)
        root  = "%s/.." % pwd
        paths = [root, pwd]

    if not isinstance(paths, list):
        paths = [paths]

    version_short  = None
    version_detail = None
    version_base   = None
    version_branch = None
    sdist_name     = None
    sdist_path     = None
    err            = ''

    # if in any of the paths a VERSION file exists, we use the detailed version
    # in there.
    for path in paths:

        try:
            version_path = "%s/VERSION" % path

            with open(version_path) as f:
                data    = f.read()
                lines   = data.split('\n', 1)
                lines   = [line.strip() for line in lines]
                lines   = [line         for line in lines if line]
                detail  = lines[-1]
                pattern = re.compile(_pat)
                match   = pattern.search(detail)

                if match:
                    version_short  = match.group('short').strip()
                    version_detail = match.group('detail').strip()
                    version_base   = match.group('base').strip()
                    version_branch = match.group('branch').strip()
                    break

        except Exception as e:
            # ignore missing VERSION file -- this is caught below.  But ew keep
            # the error message
            err += '%s\n' % repr(e)

    if version_detail:
        # check if there is also an SDIST near the version_path
        sdist_path = version_path.replace('/VERSION', '/SDIST')
        try:
            with open(sdist_path) as fh:
                sdist_name = fh.read().strip()
        except Exception as e:
            # ignore missing SDIST file
            pass

        sdist_path = version_path.replace('/VERSION', '/%s' % sdist_name)

    # check if any one worked ok
    if version_detail:
        return (version_short, version_detail, version_base, version_branch,
                sdist_name, sdist_path)
    else:
        raise RuntimeError("Cannot determine version from %s (%s)"
                          % (paths, err.strip()))


# ------------------------------------------------------------------------------





import os
import sys
import pty
import time
import errno
import shlex
import select
import signal
import termios

import threading as mt

# ------------------------------------------------------------------------------
# FIXME:
#   - child I/O should use the 'rc.connectivity' I/O base class
#   - `select()` based read/write should be added there
#


# ------------------------------------------------------------------------------
#
_CHUNKSIZE = 1024 * 1024  # default size of each read
_POLLDELAY = 0.01         # seconds in between read attempts
_DEBUG_MAX = 600


# ------------------------------------------------------------------------------
#
class PTYProcess(object):
    '''
    This class spawns a process, providing that child with pty I/O channels --
    it will maintain stdin, stdout and stderr channels to the child.  I/O
    channels are exposed as file descriptors.  The class also provides
    information on process health.  Example::
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, command, cfg=None):
        '''
        The class constructor, which runs (execvpe) command in a separately
        forked process.  The new process will inherit the environment of the
        application process.

        :type  command: string or list of strings
        :param command: The given command is what is run as a child, and
                        fed/drained via pty pipes.  If given as string,
                        command is split into an array of strings,
                        using :func:`shlex.split`.
        '''

        if isinstance(command, str):
            command = shlex.split(command)

        if not isinstance(command, list):
            raise ValueError('PTYProcess expects string or list command')

        if not command:
            raise ValueError('PTYProcess expects non-empty command')

        self.rlock       = mt.RLock()
        self.command     = command  # list of strings too run()

        self.cache       = ''       # data cache
        self.tail        = ''       # tail of data data cache for error messages
        self.child       = None     # the process as created by subprocess.Popen
        self.ptyio       = None     # the process' io channel, from pty.fork()

        self.exit_code   = None     # child died with code (may be revived)
        self.exit_signal = None     # child kill by signal (may be revived)

        self.initialize()


    # --------------------------------------------------------------------------
    #
    def __del__(self):
        '''
        Need to free pty's on destruction, otherwise we might ran out of
        them (see cat /proc/sys/kernel/pty/max)
        '''

        with self.rlock:

            try   : self.kill()
            except: pass


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        with self.rlock:

            # create the child
            self.child, self.child_fd = pty.fork()

            if not self.child:
                # child process
                os.execvpe(self.command[0], self.command, os.environ)

            else:
                # parent
                new    = termios.tcgetattr(self.child_fd)
                new[3] = new[3] & ~termios.ECHO

                termios.tcsetattr(self.child_fd, termios.TCSANOW, new)

                self.parent_in  = self.child_fd
                self.parent_out = self.child_fd


    # --------------------------------------------------------------------------
    #
    def kill(self, wstat=None):
        '''
        kill the child, close all I/O channels
        '''

        with self.rlock:

            # now we can safely kill the process
            # (unless some wait did that before)
            if wstat is None:

                if self.child:
                    # yes, we have something to kill!
                    try:
                        os.kill(self.child, signal.SIGKILL)
                    except OSError:
                        pass

                    # hey, kiddo, how did that go?
                    max_tries = 10
                    tries     =  0
                    while tries < max_tries:
                        try:
                            wpid, wstat = os.waitpid(self.child, os.WNOHANG)

                        except OSError as e:

                            # this should not have failed -- child disappeared?
                            if e.errno == errno.ECHILD:
                                self.exit_code   = None
                                self.exit_signal = None
                                wstat            = None
                                break

                            # other errors are bad, but there is not much to
                            # be done at this point
                            raise

                        if wpid:
                            break

                        time.sleep(0.1)
                        tries += 1


            # at this point, we declare the process to be gone for good
            self.child = None

            # lets see if we can perform some post-mortem analysis
            if wstat is not None:

                if os.WIFEXITED(wstat):
                    # child died of natural causes - perform autopsy...
                    self.exit_code   = os.WEXITSTATUS(wstat)
                    self.exit_signal = None

                elif os.WIFSIGNALED(wstat):
                    # murder!! Child got killed by someone!  recover evidence...
                    self.exit_code   = None
                    self.exit_signal = os.WTERMSIG(wstat)

            try:
                if self.parent_out:
                    os.close(self.parent_out)
                    self.parent_out = None

            except OSError:
                pass

          # try:
          #     if self.parent_in:
          #         os.close(self.parent_in)
          #         self.parent_in = None
          # except OSError:
          #     pass

          # try:
          #     os.close(self.parent_err)
          # except OSError:
          #     pass


    # --------------------------------------------------------------------------
    #
    def wait(self, timeout=None):
        '''
        blocks forever (or until timeout) until the child finishes on its own or
        is getting killed.
        '''

        with self.rlock:

            start = time.time()
            while True:

                if not self.alive():
                    return True

                if timeout and time.time() - start > timeout:
                    return False

                time.sleep(_POLLDELAY)


    # --------------------------------------------------------------------------
    #
    def alive(self):
        '''
        determine if the child process is still active.  If not, mark
        the child as dead and close all IO descriptors etc (:func:`kill`).
        '''

        with self.rlock:

            # do we have a child which we can check?
            if not self.child:
                return False

            wstat = None
            while True:
              # print 'waitpid %s' % self.child

                # hey, kiddo, whats up?
                try:
                    wpid, wstat = os.waitpid(self.child, os.WNOHANG)

                except OSError as e:
                    if e.errno == errno.ECHILD:
                        # child disappeared, go to zombie cleanup routine
                        break

                    raise

                # did we get a note about child termination?
                if 0 == wpid:
                    # nope, all is well - carry on
                    return True

                # Yes, we got a note.
                # Well, maybe the child fooled us and is just playing dead?
                if os.WIFSTOPPED(wstat) or os.WIFCONTINUED(wstat):
                  # print 'waitpid %s : %s - %s -- stop/cont' \
                  #         % (self.child, wpid, wstat)
                    # we don't care if someone stopped/resumed the child --
                    # that is up to higher powers.  For our purposes, the
                    # child is alive.  Ha!
                    continue

                break

            # so its dead -- make sure it stays dead, to avoid zombie
            # apocalypse...
          # print "He's dead, honeybunny, Jim is dead..."
            self.child = None
            self.kill(wstat=wstat)


    # --------------------------------------------------------------------------
    #
    def returncode(self):

        with self.rlock:
            if not self.child:
                return self.exit_code


    # --------------------------------------------------------------------------
    #
    def exit_signal(self):

        with self.rlock:
            if not self.child:
                return self.exit_signal


    # --------------------------------------------------------------------------
    #
    def read(self, size=0, timeout=0, _force=False):
        '''
        Read some data from the child.  By default, the method reads whatever is
        available on the next read, up to _CHUNKSIZE, but other read sizes can
        be specified.

        The method will return whatever data it has at timeout:

          timeout == 0: return the content of the first successful read, with
                        whatever data up to 'size' have been found.
          timeout <  0: return after first read attempt, even if no data have
                        been available.

        Cached data are used when available, thus potentially avoiging a new
        read completely or reducing the amount of data to be read.  If no cache
        is available and no no data are found after the timeout, the method
        returns an empty string (not `None`).

        This method will not fill the cache, but will read at most the requested
        amount of data.

        Note: the returned lines do *not* get '\\\\r' stripped.
        '''

        with self.rlock:

            # start the timeout timer right now.  Note that even if timeout
            # is short, and child.poll is slow, we will nevertheless attempt
            # at least one read...
            start = time.time()
            ret   = ''

            # read until we have enough data, or hit timeout ceiling...
            while True:

                # first, lets see if we still have data in the cache we
                # can return
                if len(self.cache):

                    if not size:
                        ret        = self.cache
                        self.cache = ''
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret

                    # we don't even need all of the cache
                    elif size <= len(self.cache):
                        ret        = self.cache[:size]
                        self.cache = self.cache[size:]
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret

                # otherwise we need to read some more data, right?  idle
                # wait 'til the next data chunk arrives, or 'til _POLLDELAY
                rlist, _, _ = select.select([self.parent_out], [], [],
                                             _POLLDELAY)
                # got some data?
                for f in rlist:
                    # read whatever we still need

                    readsize = _CHUNKSIZE
                    if size:
                        readsize = size - len(ret)

                    buf = os.read(f, readsize)

                    if len(buf) == 0 and sys.platform == 'darwin':
                        self.kill()
                        raise RuntimeError('unexpected EOF: %s' % self.tail)

                    self.cache += buf.replace('\r', '')


                # lets see if we still got any data in the cache we
                # can return
                if len(self.cache):

                    if not size:
                        ret        = self.cache
                        self.cache = ''
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret

                    # we don't even need all of the cache
                    elif size <= len(self.cache):
                        ret        = self.cache[:size]
                        self.cache = self.cache[size:]
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret

                # at this point, we do not have sufficient data -- only
                # return on timeout

                if timeout == 0:
                    # only return if we have data
                    if len(self.cache):
                        ret        = self.cache
                        self.cache = ''
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret

                elif timeout < 0:
                    # return of we have data or not
                    ret        = self.cache
                    self.cache = ''
                    self.tail += ret
                    self.tail  = self.tail[-256:]
                    return ret

                else:  # timeout > 0
                    # return if timeout is reached
                    now = time.time()
                    if now - start > timeout:
                        ret        = self.cache
                        self.cache = ''
                        self.tail += ret
                        self.tail  = self.tail[-256:]
                        return ret


    # --------------------------------------------------------------------------
    #
    def write(self, data):
        '''
        This method will repeatedly attempt to push the given data into the
        child's stdin pipe, until it succeeds to write all data.
        '''

        with self.rlock:

            # attempt to write forever -- until we succeeed
            while data:

                # check if the pty pipe is ready for data
                _, wlist, _ = select.select([], [self.parent_in], [],
                                            _POLLDELAY)

                for f in wlist:

                    # write will report the number of written bytes
                    size = os.write(f, data)

                    # otherwise, truncate by written data, and try again
                    data = data[size:]


# ------------------------------------------------------------------------------





import os
import time

from typing import Optional, Dict, Any, Callable

import threading       as mt
import subprocess      as sp

from .url     import Url
from .ids     import generate_id
from .shell   import sh_callout
from .logger  import Logger
from .profile import Profiler
from .modules import import_module
from .json_io import read_json, write_json
from .misc    import as_list, get_hostname, rec_makedir


# ------------------------------------------------------------------------------
#
class FluxHelper(object):

    # helper to programnatically handle flux instances and to obtain state
    # update events for flux jobs runknown in that instance.
    #
    # TODO: At the moment, that `FluxHelper` can only be used within a single
    #       process - later iterations my allow to start the service in one
    #       process and to register for callbacks etc. from another process.

    # list of allowed (reported) flux events
    event_list = {
                  'NEW',
                  'DEPEND',
                  'SCHED',
                  'RUN',
                  'CLEANUP',
                  'INACTIVE'
                 }


    # --------------------------------------------------------------------------
    #
    def __init__(self, name : str = None) -> None:

        if name: self._name = name
        else   : self._name = 'flux_helper'

        self._log  = Logger('radical.utils.%s' % self._name)
        self._prof = Profiler('radical.utils.%s' % self._name)
        self._lock = mt.RLock()

        try:
            self._mod = import_module('flux')

        except Exception:
            self._log.exception('flux import failed')
            raise

        self._flux_info   = dict()  # serializable data
        self._local_state = dict()  # non-serializable state

        self._base = '%s/%s' % (os.getcwd(), self._name)
        rec_makedir(self._base)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):

        return self._name


    # --------------------------------------------------------------------------
    #
    def start_service(self,
                      env: Optional[Dict[str,str]] = None
                     ) -> Dict[str, Any]:

        with self._lock:

            return self._locked_start_service(env)


    # --------------------------------------------------------------------------
    #
    def _locked_start_service(self,
                              env: Optional[Dict[str,str]] = None
                             ) -> Dict[str, Any]:

        flux_uid = generate_id('flux')

        check = 'flux env; echo "OK"; while true; do echo "ok"; sleep 1; done'
        start = 'flux start -o,-v,-S,log-filename=%s.log' % flux_uid
        cmd   = '/bin/bash -c "echo \\\"%s\\\" | %s"' % (check, start)

        penv  = None
        if not env:
            penv  = {k:v for k,v in os.environ.items()}
            for k,v in env.items():
                penv[k] = v

        flux_env  = dict()
        flux_term = mt.Event()
        flux_proc = sp.Popen(cmd, shell=True, env=penv,
                             stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.STDOUT)

        while flux_proc.poll() is None:

            try:
                line = flux_proc.stdout.readline()
                line = bytes.decode(line, 'utf-8').strip()

            except Exception as e:
                self._log.exception('flux service failed to start')
                raise RuntimeError('could not start flux') from e

            if not line:
                continue

            self._log.debug('%s', line)

            if line.startswith('export '):
                k, v = line.split(' ', 1)[1].strip().split('=', 1)
                flux_env[k] = v.strip('"')
                self._log.debug('%s = %s' % (k, v.strip('"')))

            elif line == 'OK':
                break

        if flux_proc.poll() is not None:
            raise RuntimeError('could not execute `flux start`')

        assert('FLUX_URI' in flux_env)

        # make sure that the flux url can be reched from other hosts
        # FIXME: this also routes local access via ssh which may slow comm
        flux_url             = Url(flux_env['FLUX_URI'])
        flux_url.host        = get_hostname()
        flux_url.schema      = 'ssh'
        flux_uri             = str(flux_url)
        flux_env['FLUX_URI'] = flux_uri

        self._prof.prof('flux_started', msg=flux_uid)

        # ----------------------------------------------------------------------
        def _watch_flux(flux_env : Dict[str,str],
                        flux_term: mt.Event
                       ) -> None:

            self._log.info('starting flux watcher')

            for k,v in flux_env.items():
                os.environ[k] = v

            while not flux_term.is_set():

                time.sleep(1)

                _, err, ret = sh_callout('flux ping -c 1 kvs')
                if ret:
                    self._log.error('flux watcher err: %s', err)
                    break

            # we only get here when the ping failed - set the event
            flux_term.set()
            self._log.warn('flux stopped')
        # ----------------------------------------------------------------------

        flux_watcher = mt.Thread(target=_watch_flux, args=[flux_env, flux_term])
        flux_watcher.daemon = True
        flux_watcher.start()

        self._log.info("flux startup successful: [%s]", flux_env['FLUX_URI'])


        self._flux_info[flux_uid]   = {'uid'      : flux_uid,
                                       'uri'      : flux_uri,
                                       'env'      : flux_env}

        self._local_state[flux_uid] = {'proc'     : flux_proc,
                                       'term'     : flux_term,
                                       'watcher'  : flux_watcher,
                                       'listeners': list(),
                                       'handles'  : list(),
                                       'executors': list(),
                                       'callbacks': list()}

        write_json('%s/%s.json' % (self._base, flux_uid),
                   self._flux_info[flux_uid])

        return self._flux_info[flux_uid]


    # --------------------------------------------------------------------------
    #
    def check_service(self, uid : str) -> Dict[str, Any]:

        with self._lock:

            if uid not in self._flux_info:
                try:
                    fname = '%s/%s.json' % (self._base, uid)
                    self._flux_info[uid] = read_json(fname)

                except Exception as e:
                    # no (valid) info file found - that service does not exist
                    raise LookupError('flux service id %s unknown' % uid) from e

            if uid not in self._local_state:
                # the service actually exists, but we don't have local state.
                # This implies that the servide was started by a different
                # process than this one.  We create a dummy local state, but
                # otherwise consider that service as valid.
                self._local_state[uid] = {'proc'     : None,
                                          'term'     : None,
                                          'watcher'  : None,
                                          'listener' : None,
                                          'handles'  : list(),
                                          'executors': list(),
                                          'callbacks': list()}

            else:
                if self._local_state[uid]['term'].is_set():
                    # the service is gone already
                    raise RuntimeError('flux service %s was terminated' % uid)


            return self._flux_info[uid]


    # --------------------------------------------------------------------------
    #
    def get_handle(self, uid : str) -> Any:

        with self._lock:

            flux_info = self.check_service(uid)
            assert(uid in self._local_state)

            try:
                handle = self._mod.Flux(url=flux_info['uri'])
                assert(handle)

            except Exception as e:
                self._log.exception('failed to create flux handle for %s' % uid)
                raise RuntimeError('failed to create flux handle') from e

            self._local_state[uid]['handles'].append(handle)

            return handle


    # --------------------------------------------------------------------------
    #
    def get_executor(self, uid : str) -> Any:

        with self._lock:

            flux_info = self.check_service(uid)
            assert(uid in self._local_state)

            try:
                args = {'url':flux_info['uri']}
                jex  = self._mod.job.executor.FluxExecutor(handle_kwargs=args)
                assert(jex)

            except Exception as e:
                self._log.exception('failed to create flux executor for %s' % uid)
                raise RuntimeError('failed to create flux executor') from e

            self._local_state[uid]['executors'].append(jex)

            return jex


    # ----------------------------------------------------------------------
    def _listen(self,
               flux_uid : str,
               flux_uri : str,
               flux_term: mt.Event,
               cb       : Callable[[str, str, float, dict], None]
              ) -> None:

        # NOTE: for services spawned in a different process, `flux_term` will be
        # `None` and `self._listen` will not be able to terminate with the
        # service.  In that case we run this thread forever - it is a daemon
        # thread and will dy with its parent process.

        handle = None
        self._log.debug('====> register for events: %s', cb)
        try:
            handle = self._mod.Flux(url=flux_uri)
            self._log.debug('=== event handle: %s', handle)
            handle.event_subscribe('job-state')
            self._log.debug('=== subscribe event')

            while flux_term is None or not flux_term.is_set():

                self._log.debug('=== recv event')

                # FIXME: how can recv be timed out or interrupted after work
                #        completed?
                event = handle.event_recv()

                if 'transitions' not in event.payload:
                    self._log.warn('unexpected flux event: %s' %
                                    event.payload)
                    continue

                transitions = as_list(event.payload['transitions'])

                for event in transitions:
                    self._log.debug('====> event: %s', event)
                    job_id, job_state, ts = event
                    if job_state not in self.event_list:
                        # we are not interested in this event
                        continue

                    with self._lock:
                        if flux_uid not in self._flux_info:
                            # service is gone
                            flux_term.set()
                            break

                        # FIXME: can we dig out exit code?  Other meta data?
                        try:
                            for cb in self._local_state[flux_uid]['callbacks']:
                                context = dict()
                                if job_state == 'INACTIVE':
                                    context = self._mod.job.event_wait(
                                            handle, job_id, "finish").context
                                cb(job_id, job_state, ts, context)
                        except:
                            self._log.exception('cb error')


        except Exception:
            self._log.exception('Error in listener loop')
            if handle:
                handle.event_unsubscribe('job-state')
                del(handle)

        finally:
            flux_term.set()

    # --------------------------------------------------------------------------
    #
    def register_callback(self,
                          uid: str,
                          cb : Callable[[str, str, float, dict], None]
                         ) -> None:

        with self._lock:

            self._log.debug('===> register cb %s', cb)

            self.check_service(uid)
            assert(uid in self._local_state)

            self._local_state[uid]['callbacks'].append(cb)

            flux_term = self._local_state[uid]['term']
            flux_uri  = self._flux_info[uid]['uri']

            flux_listener = mt.Thread(target=self._listen,
                                      args=[uid, flux_uri, flux_term, cb])
            flux_listener.daemon = True
            flux_listener.start()

            self._local_state[uid]['listeners'].append(flux_listener)




    # --------------------------------------------------------------------------
    #
    def unregister_callback(self,
                            uid: str,
                            cb : Callable[[Any], None]
                           ) -> None:

        with self._lock:

            self.check_service(uid)
            assert(uid in self._local_state)

            self._local_state[uid]['callbacks'].remove(cb)


    # --------------------------------------------------------------------------
    #
    def close_service(self, uid : str) -> None:

        with self._lock:

            self.check_service(uid)

            if self._local_state[uid]['proc'] is None:
                raise RuntimeError('cannot kill flux from this process')

            # terminate watcher and listener
            self._local_state[uid]['term'].set()

            # delete all created handles
            for handle in self._local_state[uid]['handles']:
                del(handle)

            # delete all created executors
            for jex in self._local_state[uid]['executors']:
                del(jex)

            # terminate the service process
            # FIXME: send termination signal to flux for cleanup
            self._local_state[uid]['proc'].kill()
            time.sleep(0.1)
            self._local_state[uid]['proc'].terminate()
            self._local_state[uid]['proc'].wait()

            # remove service entries
            del(self._flux_info[uid])
            del(self._local_state[uid])

            try:
                os.unlink('%s/%s.json' % (self._base, uid))
            except Exception:
                pass


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
# We provide a json based config file parser with following properties
#
#   - system config files will be merged with user configs (if those exist)
#   - python style comments are filtered out before parsing
#   - after parsing, `${ABC:default}`-style values are set or expanded via
#     `os.environ`
#   - the returned class exposes settings as dicts or attributes
#     cfg['foo']['bar'] == cfg.foo.bar
#
#
# Config Names and Locations
# --------------------------
#
# We assume two basic locations for config files: one is installed within the
# scope of a Python module, the other one is under user control, and usually in
# the users home directory.  The  config reader accepts the following parameters
# to derive the exact locations:
#
#   - module: name of module under which the config is installed
#   - path  : config file path relative to the module home
#   - name  : config file name relative to the path
#
# The `module` string is interpreted as follows:
#
#   module         = 'module'
#   module_path    = radical.utils.debug.find_module(module)
#   usr_base_dir   = os.environ.get('RADICAL_CONFIG_USER_DIR') or \
#                    os.environ.get('HOME', '/tmp')
#   sys_config_dir = '%s/configs'     % (module_path)
#   usr_config_dir = '%s/.%s/configs' % (usr_base_dir, module.replace('.', '/'))
#
# so the location of the module's `__init__.py` is used to derive the location
# of the installed system config files, and the module name is used to derive
# the location of the user provided config files.
#
# For example, the module `radical.utils` will have the following config dirs:
#
#   sys_config_dir = /tmp/ve/lib/python2.7/site-packages/radical/utils/configs
#   usr_config_dir = /home/merzky/.radical/utils/configs
#
# After loading the system level config file, any existing user level config
# file is merged into it, via
#
#   radical.utils.dict_merge(user_cfg, system_cfg, mode='overwrite')
#
# so that the user config settings supersede the system config settings.
#
# Both path and name specifiers can contain `*` as wildcard, which is then
# interpreted as by `glob()`.  If that wildcard exist, then all matching config
# files are read into *one* configuration dict, where each root key is set to
# the value the '*' expands to (minus the `.json` extension).
#
# For example, the name `radical.pilot.resource_*` with the following config
# files:
#
#   /tmp/ve/[...]/radical/pilot/configs/resource_xsede.json
#   /tmp/ve/[...]/radical/pilot/configs/resource_ncsa.json
#
# will result in a config dict like:
#
#   {
#     'xsede' : { 'foo' : 'bar' },
#     'ncsa'  : { 'fiz' : 'baz' }
#   }
#
#
# Queries
# -------
#
# We support two types of queries on the resulting parsed configs:
#
#   - dict like queries (via `ru.DictMixin`)
#   - the `query(key)` method returns a single value, or 'None' if not found.
#
# In the latter `query()` case, the `key` can be specified as dot-separated
# path, so that the following two snippets are equivalent (assuming that a
# `foo.bar` section exists):
#
#   val = cfg['foo']['bar'].get('baz')
#   val = cfg.query('foo.bar.baz')
#
#
# Environment
# -----------
#
# Towards `os.environ` completion, we support the following syntax in all string
# *values* (not keys):
#
#   '${RADICAL_UTILS_ENV:default_value}'
#
# which will be replaced by
#
#   `os.environ.get('RADICAL_UTILS_ENV', 'default_value')`
#
# The default value is optional, an empty string is used if no default value is
# given.  Env evaluation is only performed at time of parsing, not at time of
# query.  RU attempts to convert env variables to float and int - if that fails,
# values are stored as strings.
#
#
# Validation
# ----------
#
# It probably makes sense to switch to a json schema validator at some point,
# see for example https://pypi.python.org/pypi/json-schema-validator. For now
# this implementation remains schema-less, and will thus, in a very pythonesque
# way, only fail once values are queried or used.
#
#
# Implementation
# --------------
#
# This implementation is based on typed dictionaries which are accessed as
# `munch`'ed object hierarchy.
#
# ------------------------------------------------------------------------------

import os
import glob

from .debug      import find_module
from .misc       import expand_env as ru_expand_env
from .json_io    import read_json
from .dict_mixin import dict_merge
from .munch      import Munch

from .singleton  import Singleton


# ------------------------------------------------------------------------------
#
class Config(Munch):

    _self_default = True

    # FIXME: we should cache config files after reading, so that repeated
    #        instance creations do not trigger a new (identical) round of
    #        parsing.

    # --------------------------------------------------------------------------
    #
    def __init__(self, module=None, category=None, name=None, cfg=None,
                       from_dict=None, path=None, expand=True, env=None,
                       _internal=False):
        '''
        Load a config (json) file from the module's config tree, and overload
        any user specific config settings if found.

        module:    used to determine the module's config file location
                   - default: `radical.utils`
        category:  name of config to be loaded from module's config path
        name:      specify a specific configuration to be used
        path:      path to app config json to be used for initialization
        cfg:       application config dict to be used for initialization
        from_dict: alias for cfg, to satisfy base class constructor
        expand:    enable / disable environment var expansion
                   - default: True
        env:       environment dictionary to be used for expansion
                   - default: `os.environ`

        The naming of config files follows this rule:

          `<category>_<name>.json`

        For example, if the following is used in a system python installation:

            ru.Config('radical.pilot', category='session', name='minimal')

        it would attempt to load (depending on system details):

            /usr/lib/python3/site-packages/\
                 radical/pilot/configs/session_mininmal.json

        NOTE: Keys containing an underscore are not exposed via the API.
              Keys containing dots are split and interpreted as paths in the
              configuration hierarchy.
        '''

        if from_dict:
            # if we could only overload constructors by signature... :-/
            # As it is, we have to emulate that...
            assert(not module   and
                   not category and
                   not name     and
                   not cfg      and
                   not path     and
                   not env      and
                   not _internal), 'from_dict must be exclusive'
            super().__init__(from_dict=from_dict)
            return

        if path and cfg:
            raise ValueError('conflicting initializers (path, cfg)')

        if path:
            cfg = read_json(path)

        if not cfg:
            # just use config files
            cfg = dict()

        # if a category has dot limited elements and no module is given,
        # interpret the first part as module
        # radical.pilot.session -> [radical.pilot, session]
        if not module and category and '.' in category:
            elems    = category.split('.')
            module   = '.'.join(elems[:-1])
            category = elems[-1]

        # do the same if module is given but not the category
        if not category and module and '.' in module:
            elems    = module.split('.')
            module   = '.'.join(elems[:-1])
            category = elems[-1]

        # if a category starts with a module prefix, strip that prefix
        if category and category.startswith('%s.' % module):
            category = category[len(module) + 1:]

        name_orig = name
        if not name:
            # by default, load the default config
            name = 'default'

        # purge module and category parts from name
        if module and name.startswith(module + '.'):
            name = name[len(module) + 1:]

        if category and name.startswith(category + '.'):
            name = name[len(category) + 1:]

        if name.startswith('/'):

            # load config from abs path
            sys_fspec = name
            usr_fspec = None

            if '*' in name: starred = True
            else          : starred = False

        elif module and category:

            # construct cfg file paths
            modpath = find_module(module)
            if not modpath:
                raise ValueError("Cannot find module %s" % module)

            home    = os.environ.get('HOME', '/tmp')
            home    = os.environ.get('RADICAL_CONFIG_USER_DIR', home)
            sys_dir = "%s/configs"     % (modpath)
            usr_dir = "%s/.%s/configs" % (home, module.replace('.', '/'))
            fname   = '%s_%s.json'     % (category.replace('.', '/'), name)

            sys_fspec = '%s/%s' % (sys_dir, fname)
            usr_fspec = '%s/%s' % (usr_dir, fname)

            if '*' in fname: starred = True
            else           : starred = False

        else:
            # we can't load a config file - just use the app config
            sys_fspec = None
            usr_fspec = None
            starred   = False

        sys_cfg = dict()
        usr_cfg = dict()
        app_cfg = cfg

        if _internal:
            # no need to look at the FS, just convert the given cfg dict
            sys_fspec = None
            usr_fspec = None

        nfiles = 0
        if not starred:

            if sys_fspec:

                sys_fname = sys_fspec

                if not os.path.isfile(sys_fname):
                    sys_fname += '.json'

                if os.path.isfile(sys_fname):
                    sys_cfg = read_json(sys_fname)
                    nfiles += 1

            if usr_fspec:

                usr_fname = usr_fspec

                if not os.path.isfile(usr_fname):
                    usr_fname += '.json'

                if os.path.isfile(usr_fname):
                    usr_cfg = read_json(usr_fname)
                    nfiles += 1

        else:

            # wildcard mode: whatever the '*' expands into is used as root dict
            # entry, and the respective content of the config file is stored
            # underneath it.
            if sys_fspec:

                postfix_len = len('.json')                      # ' .json'
                prefix_len  = len(sys_fspec) - postfix_len - 1  # '*.json'

                for sys_fname in glob.glob(sys_fspec):

                    base = sys_fname[prefix_len:-postfix_len]
                    scfg = read_json(sys_fname)
                    sys_cfg[base] = scfg
                    nfiles += 1

            if usr_fspec:

                postfix_len = len('.json')                      # ' .json'
                prefix_len  = len(usr_fspec) - postfix_len - 1  # '*.json'

                for usr_fname in glob.glob(usr_fspec):

                    base = usr_fname[prefix_len:-postfix_len]
                    ucfg = read_json(usr_fname)
                    usr_cfg[base] = ucfg
                    nfiles += 1

        # if we did not find *any* file, and the original `name` was None,
        # then try to load config files w/o name
        # Example: if there is no `registry_default.json`, then try to load
        # `registry.json`.
        if nfiles == 0 and name_orig is None and not _internal and category:

            fname     = '%s.json' % (category.replace('.', '/'))
            sys_fname = '%s/%s'   % (sys_dir, fname)
            usr_fname = '%s/%s'   % (usr_dir, fname)

            if os.path.isfile(sys_fname): sys_cfg = read_json(sys_fname)
            if os.path.isfile(usr_fname): usr_cfg = read_json(usr_fname)


        # merge sys, usr and app cfg before expansion
        cfg_dict = dict()
        cfg_dict = dict_merge(cfg_dict, sys_cfg, policy='overwrite')
        cfg_dict = dict_merge(cfg_dict, usr_cfg, policy='overwrite')
        cfg_dict = dict_merge(cfg_dict, app_cfg, policy='overwrite')

        if expand:
            cfg_dict = ru_expand_env(cfg_dict, env=env)

        super().__init__(from_dict=cfg_dict)


# ------------------------------------------------------------------------------
#
class DefaultConfig(Config, metaclass=Singleton):
    '''
    The settings in this default config are, unsurprisingly, used as default
    values for various RU classes and methods, as for example for log file
    locations, log levels, profile locations, etc.
    '''

    _schema   = {
                   'log_lvl'    : str,
                   'log_tgt'    : str,
                   'log_dir'    : str,
                   'report'     : bool,
                   'report_tgt' : str,
                   'report_dir' : str,
                   'profile'    : bool,
                   'profile_dir': str,
                 }


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        super(DefaultConfig, self).__init__(module='radical.utils.utils',
                                            name='default')


# ------------------------------------------------------------------------------






import os
import inspect
import pkgutil

import importlib.util


_id_cnt = 0


# ------------------------------------------------------------------------------
#
# to keep RU 2.6 compatible, we provide import_module which works around some
# quirks of __import__ when being used with dotted names. This is what the
# python docs recommend to use.  This basically steps down the module path and
# loads the respective submodule until arriving at the target.
#
def import_module(name):

    mod = __import__(name)
    for s in name.split('.')[1:]:
        mod = getattr(mod, s)
    return mod


# ------------------------------------------------------------------------------
#
# as import_module, but without the import part :-P
#
def find_module(name):

    package = pkgutil.get_loader(name)

    if not package:
        return None

    if '_NamespaceLoader' in str(package):
        # since Python 3.5, loaders differ between modules and namespaces
        return package._path._path[0]                    # pylint: disable=W0212
    else:
        return os.path.dirname(package.get_filename())


# ------------------------------------------------------------------------------
#
# a helper to load functions and classes from user provided source file which
# are *not* installed as modules.  All symbols from that file are loaded, and
# returned is a dictionary with the following structure:
#
# symbols = {'classes'  : {'Foo': <class 'mod_0001.Foo'>,
#                          'Bar': <class 'mod_0001.Bar'>,
#                          ...
#                         },
#            'functions': {'foo': <function foo at 0x7f532d241d40>,
#                          'bar': <function bar at 0x7f532d241d40>,
#                          ...
#                         }
#           }
#
def import_file(path):

    global _id_cnt
    _id_cnt += 1

    uid  = 'mod_%d' % _id_cnt
    spec = importlib.util.spec_from_file_location(uid, path)
    mod  = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(mod)

    symbols = {'functions': dict(),
               'classes'  : dict()}

    for k,v in mod.__dict__.items():
        if not k.startswith('__'):
            if inspect.isclass(v):    symbols['classes'  ][k] = v
            if inspect.isfunction(v): symbols['functions'][k] = v

    return symbols


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pylint: disable=protected-access

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import time

import radical.utils as ru


# ------------------------------------------------------------------------------
#
# helper state
#
_state = 0


# ------------------------------------------------------------------------------
def test_object_cache():
    """
    Test object cache
    """

    global _state                                        # pylint: disable=W0603

    # --------------------------------------------------------------------------
    class _Test(object):

        def __init__(self):

            global _state                                # pylint: disable=W0603
            _state += 1


        def __del__(self):

            global _state                                # pylint: disable=W0603
            _state -= 1

    # --------------------------------------------------------------------------
    def _test_1(oc):

        # create/cache two instances
        t1_1 = oc.get_obj('id_1', _Test); assert(_state == 1), "%d" % _state
        t1_2 = oc.get_obj('id_1', _Test); assert(_state == 1), "%d" % _state
        t2_1 = oc.get_obj('id_2', _Test); assert(_state == 2), "%d" % _state
        t2_2 = oc.get_obj('id_2', _Test); assert(_state == 2), "%d" % _state

        # we now have two different instances
        assert(t1_1 == t1_2)
        assert(t2_1 == t2_2)
        assert(t1_1 != t2_1)

        # one remove does not delete first instance
        assert(oc.rem_obj(t1_1))
        time.sleep(0.2)
        assert(_state == 2), "%d" % _state

        # second remove triggers *delayed* delete
        assert(oc.rem_obj(t1_2))
        assert(_state == 2), "%d" % _state

        # instances gow out of scope here - but still live in the object_cache
        return str(t2_1)

    # --------------------------------------------------------------------------
    def _test_2(oc, uid):

        t2_3 = oc.get_obj('id_2', _Test)
        assert(_state == 1), "%d" % _state
        assert(str(t2_3) == uid)

        assert(oc.rem_obj(t2_3))  # for t2_1
        assert(oc.rem_obj(t2_3))  # for t2_2
        assert(oc.rem_obj(t2_3))  # for t2_3
    # --------------------------------------------------------------------------

    # initial state
    assert(_state == 0), "%d" % _state
    oc  = ru.ObjectCache(timeout=0.1)
    uid = _test_1(oc)

    time.sleep(0.2)
    assert(_state == 1), "%d" % _state

    _test_2(oc, uid)

    assert(_state == 1), "%d" % _state
    time.sleep(0.2)
    assert(_state == 0), "%d" % _state

    assert(oc.rem_obj('foo'              ) is False)
    assert(oc.rem_obj('foo', ns='unknown') is False)

    # test removal w/o timeout (new singleton instance)
    del(ru.Singleton._instances[ru.ObjectCache])
    oc = ru.ObjectCache(timeout=0.0)
    x  = oc.get_obj('no_timeout', _Test)
    oc.rem_obj(x)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_object_cache()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import time

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def _cmd(cmd):

    _, _, ret = ru.sh_callout(cmd)

    if ret == 0:
        return True
    else:
      # print 'cmd: %s' % cmd
      # print 'out: %s' % out
      # print 'err: %s' % err
        return False


# ------------------------------------------------------------------------------
#
def test_reporter():

    pname = 'ru.%d'        % os.getpid()
    fname = '/tmp/%s.prof' % pname

    os.environ['RADICAL_UTILS_REPORT']     = 'True'
    os.environ['RADICAL_UTILS_REPORT_TGT'] = fname

    rep = ru.Reporter(name=pname, ns='radical.utils')

    rep.header  ('header  \n')
    rep.info    ('info    \n')
    rep.progress('progress\n')
    rep.ok      ('ok      \n')
    rep.warn    ('warn    \n')
    rep.error   ('error   \n')
    rep.plain   ('plain   \n')


    rep.info('test idler:')
    rep.idle(mode='start')
    for _ in range(3):
        rep.idle()
        time.sleep(0.03)
    rep.idle(color='ok', c='.')
    rep.idle(color='error', c='.')
    for _ in range(3):
        rep.idle()
        time.sleep(0.01)

    rep.idle(mode='stop')
    rep.ok('>>done\n')


    # pylint disable=E0501
    rep.info('idle test\n')
    rep.info('1234567891         2         3         4         5         6         7         8\n\t')  # noqa
    rep.info('.0.........0.........0.........0.........0.........0.........0.........0')              # noqa
    # pylint enable=E0501

    rep.idle(mode='start')
    for _ in range(200):
        rep.idle() ; time.sleep(0.01)
        rep.idle() ; time.sleep(0.01)
        rep.idle() ; time.sleep(0.01)
        rep.idle() ; time.sleep(0.01)
        rep.idle(color='ok', c="+")
    rep.idle(mode='stop')

    rep.set_style('error', color='yellow', style='ELTTTTMELE', segment='X')
    rep.error('error')

    try                  : rep.exit('exit', 1)
    except SystemExit    : assert(True)
    except Exception as e: assert(False), 'expected system exit, got %s' % e

    assert(os.path.isfile(fname))
    assert(_cmd('grep -e "header"    %s' % fname))

    try   : os.unlink(fname)
    except: pass


# ------------------------------------------------------------------------------
#
def test_env():
    '''
    Print out some messages with different log levels
    '''

    def _assert_reporter(pname, fname, val=True):

        rep  = ru.Reporter(name=pname, ns='radical.utils.test', path='/tmp/')
        rep.info('foo')

        if fname:
            assert(val == os.path.isfile(fname))
            assert(val == _cmd('grep -e "foo" %s' % fname))

            try   : os.unlink(fname)
            except: pass



    for key in ['RADICAL_REPORT',
                'RADICAL_UTILS_REPORT',
                'RADICAL_UTILS_TEST_REPORT']:

        for k in list(os.environ.keys()):
            if k.startswith('RADICAL'):
                del(os.environ[k])

        pname = 'ru.%d'        % os.getpid()
        fname = '/tmp/%s.prof' % pname
        _assert_reporter(pname, fname, False)

        os.environ['RADICAL_UTILS_REPORT_TGT'] = fname

        os.environ[key] = ''     ;  _assert_reporter(pname, fname)
        os.environ[key] = '0'    ;  _assert_reporter(pname, None)
        os.environ[key] = '1'    ;  _assert_reporter(pname, fname)
        os.environ[key] = 'True' ;  _assert_reporter(pname, fname)
        os.environ[key] = 'False';  _assert_reporter(pname, None)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_env()
    test_reporter()


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

import zmq
import msgpack

import threading as mt

from ..atfork  import atfork
from ..config  import Config
from ..ids     import generate_id, ID_CUSTOM
from ..url     import Url
from ..misc    import get_hostip, is_string, as_string, as_bytes, as_list, noop
from ..logger  import Logger
from ..profile import Profiler

from .bridge   import Bridge
from .utils    import no_intr, log_bulk


# ------------------------------------------------------------------------------
#
_LINGER_TIMEOUT  =   250  # ms to linger after close
_HIGH_WATER_MARK =     0  # number of messages to buffer before dropping
                          # 0:  infinite


# ------------------------------------------------------------------------------
#
def _atfork_child():
    Subscriber._callbacks = dict()                                        # noqa


atfork(noop, noop, _atfork_child)


# ------------------------------------------------------------------------------
#
# Notifications between components are based on pubsub channels.  Those channels
# have different scope (bound to the channel name).  Only one specific topic is
# predefined: 'state' will be used for unit state updates.
#
class PubSub(Bridge):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg=None, channel=None):

        if cfg and not channel and is_string(cfg):
            # allow construction with only channel name
            channel = cfg
            cfg     = None

        if   cfg    : cfg = Config(cfg=cfg)
        elif channel: cfg = Config(cfg={'channel': channel})
        else: raise RuntimeError('PubSub needs cfg or channel parameter')

        if not cfg.channel:
            raise ValueError('no channel name provided for pubsub')

        if not cfg.uid:
            cfg.uid = generate_id('%s.bridge.%%(counter)04d' % cfg.channel,
                                  ID_CUSTOM)

        super(PubSub, self).__init__(cfg)


    # --------------------------------------------------------------------------
    #
    # protocol independent addr query
    @property
    def type_in(self):
        return 'pub'

    @property
    def type_out(self):
        return 'sub'

    @property
    def addr_in(self):
        return self.addr_pub

    @property
    def addr_out(self):
        return self.addr_sub

    # protocol dependent addr query
    @property
    def addr_pub(self):
        return self._addr_pub

    @property
    def addr_sub(self):
        return self._addr_sub


    # --------------------------------------------------------------------------
    #
    def _bridge_initialize(self):

        self._log.info('initialize bridge %s', self._uid)

        self._url        = 'tcp://*:*'
        self._lock       = mt.Lock()

        self._ctx        = zmq.Context()  # rely on GC for destruction
        self._pub        = self._ctx.socket(zmq.XSUB)
        self._pub.linger = _LINGER_TIMEOUT
        self._pub.hwm    = _HIGH_WATER_MARK
        self._pub.bind(self._url)

        self._sub        = self._ctx.socket(zmq.XPUB)
        self._sub.linger = _LINGER_TIMEOUT
        self._sub.hwm    = _HIGH_WATER_MARK
        self._sub.bind(self._url)

        # communicate the bridge ports to the parent process
        _addr_pub = as_string(self._pub.getsockopt(zmq.LAST_ENDPOINT))
        _addr_sub = as_string(self._sub.getsockopt(zmq.LAST_ENDPOINT))

        # store addresses
        self._addr_pub = Url(_addr_pub)
        self._addr_sub = Url(_addr_sub)

        # use the local hostip for bridge addresses
        self._addr_pub.host = get_hostip()
        self._addr_sub.host = get_hostip()

        self._log.info('bridge pub on  %s: %s'  % (self._uid, self._addr_pub))
        self._log.info('       sub on  %s: %s'  % (self._uid, self._addr_sub))

        # start polling for messages
        self._poll = zmq.Poller()
        self._poll.register(self._pub, zmq.POLLIN)
        self._poll.register(self._sub, zmq.POLLIN)


    # --------------------------------------------------------------------------
    #
    def _bridge_work(self):

        # we could use a zmq proxy - but we rather code it directly to have
        # proper logging, timing, etc.  But the code for the proxy would be:
        #
        #     zmq.proxy(socket_pub, socket_sub)
        #
        # That's the equivalent of the code below.

        while not self._term.is_set():

            # timeout in ms
            socks = dict(self._poll.poll(timeout=500))

            if self._sub in socks:

                # if the sub socket signals a message, it's likely
                # a topic subscription.  Forward that to the pub
                # channel, so the bridge subscribes for the respective
                # message topic.
                msg = self._sub.recv()
                self._pub.send(msg)

                self._prof.prof('subscribe', uid=self._uid, msg=msg)
              # log_bulk(self._log, msg, '~~ %s' % self.channel)


            if self._pub in socks:

                # if the pub socket signals a message, get the message
                # and forward it to the sub channel, no questions asked.
                msg = self._pub.recv()
                self._sub.send(msg)

              # self._prof.prof('msg_fwd', uid=self._uid, msg=msg)
              # log_bulk(self._log, msg, '<> %s' % self.channel)


# ------------------------------------------------------------------------------
#
class Publisher(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, log=None, prof=None, path=None):

        self._channel  = channel
        self._url      = as_string(url)
        self._log      = log
        self._prof     = prof
        self._lock     = mt.Lock()

        # FIXME: no uid ns
        self._uid      = generate_id('%s.pub.%s' % (self._channel,
                                                   '%(counter)04d'), ID_CUSTOM)

        if not self._url:
            self._url = Bridge.get_config(channel, path).pub

        if not log:
            self._log  = Logger(name=self._uid, ns='radical.utils.zmq')

        if not prof:
            self._prof = Profiler(name=self._uid, ns='radical.utils.zmq')
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect pub to %s: %s'  % (self._channel, self._url))

        self._ctx           = zmq.Context()  # rely on GC for destruction
        self._socket        = self._ctx.socket(zmq.PUB)
        self._socket.linger = _LINGER_TIMEOUT
        self._socket.hwm    = _HIGH_WATER_MARK
        self._socket.connect(self._url)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def put(self, topic, msg):

        assert(isinstance(topic, str )), 'invalid topic type'

      # self._log.debug('=== put %s : %s: %s', topic, self.channel, msg)
      # self._prof.prof('put', uid=self._uid, msg=msg)
      # log_bulk(self._log, msg, '-> %s' % self.channel)

        btopic = as_bytes(topic.replace(' ', '_'))
        bmsg   = msgpack.packb(msg)
        data   = btopic + b' ' + bmsg

        self._socket.send(data)


# ------------------------------------------------------------------------------
#
class Subscriber(object):

    # instead of creating a new listener thread for each endpoint which then, on
    # incoming messages, calls a subscriber callback, we only create *one*
    # listening thread per ZMQ endpoint address and call *all* registered
    # callbacks in that thread.  We hold those endpoints in a class dict, so
    # that all class instances share that information
    _callbacks = dict()


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_nowait(socket, lock, timeout, log, prof):

        # FIXME: add logging

        if socket.poll(flags=zmq.POLLIN, timeout=timeout):

            raw         = no_intr(socket.recv, flags=zmq.NOBLOCK)
            topic, data = raw.split(b' ', 1)
            msg         = msgpack.unpackb(data)

            return [as_string(topic), as_string(msg)]

        return None, None


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _listener(url, log, prof):

      # assert(url in Subscriber._callbacks)

        try:
          # uid    = Subscriber._callbacks.get(url, {}).get('uid')
            lock   = Subscriber._callbacks.get(url, {}).get('lock')
            term   = Subscriber._callbacks.get(url, {}).get('term')
            socket = Subscriber._callbacks.get(url, {}).get('socket')

            while not term.is_set():

                # this list is dynamic
                callbacks  = Subscriber._callbacks[url]['callbacks']
                topic, msg = Subscriber._get_nowait(socket, lock, 500, log, prof)

                if topic:
                    for cb, _lock in callbacks:
                      # prof.prof('call_cb', uid=uid, msg=cb.__name__)
                        if _lock:
                            with _lock:
                                cb(topic, msg)
                        else:
                            cb(topic, msg)
        except:
            log.exception('listener died')


    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, topic=None, cb=None,
                                log=None, prof=None, path=None):
        '''
        If a `topic` is given, the channel will subscribe to that topic
        immediately.

        When a callback `cb` is specified, then the Subscriber c'tor will spawn
        a separate thread which continues to listen on the channel, and the cb
        is invoked on any incoming message.  The topic will be the first, the
        message will be the second argument to the cb.
        '''

        self._channel  = channel
        self._url      = as_string(url)
        self._topic    = as_list(topic)
        self._cb       = cb
        self._log      = log
        self._prof     = prof

        self._uid      = generate_id('%s.sub.%s' % (self._channel,
                                                   '%(counter)04d'), ID_CUSTOM)

        if not self._url:
            self._url = Bridge.get_config(channel, path).sub

        if not self._log:
            self._log = Logger(name=self._uid, ns='radical.utils.zmq')

        if not self._prof:
            self._prof = Profiler(name=self._uid, ns='radical.utils.zmq')
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect sub to %s: %s'  % (self._channel, self._url))

        self._lock     = mt.Lock()
        self._ctx      = zmq.Context()  # rely on GC for destruction

        if self._url not in Subscriber._callbacks:

            s        = self._ctx.socket(zmq.SUB)
            s.linger = _LINGER_TIMEOUT
            s.hwm    = _HIGH_WATER_MARK
            s.connect(self._url)

            Subscriber._callbacks[self._url] = {'uid'      : self._uid,
                                                'socket'   : s,
                                                'channel'  : channel,
                                                'lock'     : mt.Lock(),
                                                'term'     : mt.Event(),
                                                'thread'   : None,
                                                'callbacks': list()}

        # only allow `get()` and `get_nowait()`
        self._interactive = True

        if topic:
            self.subscribe(topic, cb)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def _start_listener(self):

        # only start if needed
        if Subscriber._callbacks[self._url]['thread']:
            return

        t = mt.Thread(target=Subscriber._listener,
                      args=[self._url, self._log, self._prof])
        t.daemon = True
        t.start()

        Subscriber._callbacks[self._url]['thread'] = t


    # --------------------------------------------------------------------------
    #
    def _stop_listener(self, force=False):

        # only stop listener if no callbacks remain registered (unless forced)
        if force or not Subscriber._callbacks[self._url]['callbacks']:
            if  Subscriber._callbacks[self._url]['thread']:
                Subscriber._callbacks[self._url]['term'  ].set()
                Subscriber._callbacks[self._url]['thread'].join()
                Subscriber._callbacks[self._url]['term'  ].unset()
                Subscriber._callbacks[self._url]['thread'] = None


    # --------------------------------------------------------------------------
    #
    def subscribe(self, topic, cb=None, lock=None):

        # if we need to serve callbacks, then open a thread to watch the socket
        # and register the callbacks.  If a thread is already runnning on that
        # channel, just register the callback.
        #
        # Note that once a thread is watching a socket, we cannot allow to use
        # `get()` and `get_nowait()` anymore, as those will interfere with the
        # thread consuming the messages,
        #
        # The given lock (if any) is used to shield concurrent cb invokations.

        if cb:

            self._interactive = False
            self._start_listener()
            Subscriber._callbacks[self._url]['callbacks'].append([cb, lock])

        sock  = Subscriber._callbacks[self._url]['socket']
        topic = str(topic).replace(' ', '_')
        log_bulk(self._log, topic, '~~ %s' % self.channel)

        with self._lock:
            no_intr(sock.setsockopt, zmq.SUBSCRIBE, as_bytes(topic))


    # --------------------------------------------------------------------------
    #
    def unsubscribe(self, cb):

        if self._url in Subscriber._callbacks:
            for _cb, _lock in Subscriber._callbacks[self._url]['callbacks']:
                if cb == _cb:
                    Subscriber._callbacks[self._url]['callbacks'].remove([_cb, _lock])
                    break

        self._stop_listener()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._stop_listener(force=True)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._interactive:
            raise RuntimeError('invalid get(): callbacks are registered')


        # FIXME: add timeout to allow for graceful termination
        #
        sock = Subscriber._callbacks[self._url]['socket']

        with self._lock:
            raw = no_intr(sock.recv)

        topic, data = raw.split(b' ', 1)
        msg = msgpack.unpackb(data)

        log_bulk(self._log, msg, '<- %s' % self.channel)

        return [as_string(topic), as_string(msg)]


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout=None):

        # FIXME:  does this duplicate _get_nowait? why / why not?

        if not self._interactive:
            raise RuntimeError('invalid get_nowait(): callbacks are registered')

        sock = Subscriber._callbacks[self._url]['socket']

        if no_intr(sock.poll, flags=zmq.POLLIN, timeout=timeout):

            with self._lock:
                raw = no_intr(sock.recv, flags=zmq.NOBLOCK)

            topic, data = raw.split(b' ', 1)
            msg = msgpack.unpackb(data)

            log_bulk(self._log, msg, '<- %s' % self.channel)

            return [as_string(topic), as_string(msg)]

        else:
            return [None, None]


# ------------------------------------------------------------------------------





import re
import os
import hashlib
import tempfile

from .misc  import as_list
from .shell import sh_callout


# we know that some env vars are not worth preserving.  We explicitly exclude
# those which are common to have complex syntax and need serious caution on
# shell escaping:
BLACKLIST  = ['PS1', 'LS_COLORS', '_', 'SHLVL']

# Identical task `pre_exec_cached` settings will result in the same environment
# settings, so we cache those environments here.  We rely on a hash to ensure
# `pre_exec_cached` identity.  Note that this assumes that settings do not
# depend on, say, the unit ID or similar, which needs very clear and prominent
# documentation.  Caching can be turned off by adding a unique noop string to
# the `pre_exec_cached` list - but we probably also add a config flag if that
# becomes a common issue.
_env_cache = dict()


# ------------------------------------------------------------------------------
#
def env_read(fname):
    '''
    helper to parse environment from a file: this method parses the output of
    `env` and returns a dict with the found environment settings.
    '''

    with open(fname, 'r') as fin:
        lines = fin.readlines()

    return env_read_lines(lines)


# ------------------------------------------------------------------------------
#
def env_read_lines(lines):

    # POSIX definition of variable names
    key_pat = r'^[A-Za-z_][A-Za-z_0-9]*$'
    env     = dict()
    key     = None
    val     = ''

    for line in lines:

        # remove newline
        line = line.rstrip('\n')

        if not line:
            continue

        # search for new key
        if '=' not in line:
            # no key present - append linebreak and line to value
            val += '\n'
            val += line
            continue


        this_key, this_val = line.split('=', 1)

        if re.match(key_pat, this_key):
            # valid key - store previous key/val if we have any, and
            # initialize `key` and `val`
            if key and key not in BLACKLIST:
                env[key] = val

            key = this_key
            val = this_val
        else:
            # invalid key - append linebreak and line to value
            val += '\n'
            val += line

    # store last key/val if we have any
    if key and key not in BLACKLIST:
        env[key] = val

    return env


# ------------------------------------------------------------------------------
#
def _quote(data):

    if "'" in data or '$' in data or '`' in data:
        # cannot use single quote, so use double quote and escale all other
        # double quotes in the data
        # NOTE: we only support these three types of shell directives
        data = '"' + data.replace('"', '\\"') + '"'

    else:
        # single quotes will do
        data = "'" + data +  "'"

    return data


# ------------------------------------------------------------------------------
#
def _unquote(data):

    if data.startswith("'") and data.endswith("'"):
        # just remove enclosing single quotes - no nesting
        data = data[1:-1]

    elif data.startswith('"') and data.endswith('"'):
        # remove enclosing double quotes, and replace all occurences of escaled
        # double quotes (`\"`) with an unescaled one (`"`).
        data = data[1:-1]
        data = data.replace('\\"', '"')

    return data


# ------------------------------------------------------------------------------
#
def env_eval(fname):
    '''
    helper to create a dictionary with the env settings in the specified file
    which contains `unset` and `export` directives, or simple 'key=val' lines
    '''

    env = dict()
    with open(fname, 'r') as fin:

        for line in fin.readlines():

            line = line.strip()

            if not line:
                continue

            if line.startswith('#'):
                continue

            if line.startswith('unset ') :
                _, spec = line.split(' ', 1)
                k = spec.strip()
                if k not in env:
                    continue
                del(env[k])

            elif line.startswith('export ') :
                _, spec = line.split(' ', 1)
                k,v = spec.split('=', 1)
                env[k] = _unquote(v)

            else:
                k,v = line.split('=', 1)
                env[k] = _unquote(v)

    return env


# ------------------------------------------------------------------------------
#
def env_prep(environment=None, unset=None, pre_exec=None,
             pre_exec_cached=None, script_path=None):
    '''
    Create a shell script which restores the environment specified in
    `environment`
    environment (dict).  While doing so, ensure that all env variables *not*
    defined in `environment` but defined in `unset` (list) are unset.  Also ensure
    that all commands provided in `pre_exec_cached` (list) are executed after
    these settings.

    Once the shell script is created, run it and dump the resulting env, then
    read it back via `env_read()` and return the resulting env dict - that
    can then be used for process fork/execve to run a process is the thus
    defined environment.

    The resulting environment will be cached: a subsequent call with the same
    set of parameters will simply return a previously cached environment if it
    exists.

    If `script_path` is given, a shell script will be created in the given
    location so that shell commands can source it and restore the specified
    environment.

    Any commands given in 'pre_exec' will be part of the cached script, and will
    thus *not* be executed when preparing the env, but *will* be executed
    whenever the prepared shell script is sources.  The returned env dictionary
    will thus *not* include the effects of those injected commands.
    '''

    global _env_cache

    # defaults
    if environment     is None: environment     = os.environ
    if unset           is None: unset           = list()
    if pre_exec        is None: pre_exec        = list()
    if pre_exec_cached is None: pre_exec_cached = list()

    if pre_exec and not script_path:
        raise ValueError('`pre_exec` must be used with `script_path`')

    # empty `pre_exec*` settings are ok - just ensure correct type
    pre_exec        = as_list(pre_exec       )
    pre_exec_cached = as_list(pre_exec_cached)

    # cache lookup
    cache_key = str(sorted(environment.items())) \
              + str(sorted(unset))               \
              + str(sorted(pre_exec))            \
              + str(sorted(pre_exec_cached))
    cache_md5 = hashlib.md5(cache_key.encode('utf-8')).hexdigest()

    if cache_md5 in _env_cache:
        env = _env_cache[cache_md5]

    else:
        # cache miss

        # Write a temporary shell script which
        #
        #   - unsets all variables which are not defined in `environment`
        #     but are defined in the `unset` list;
        #   - unset all blacklisted vars;
        #   - sets all variables defined in the `environment` dict;
        #   - inserts all the `pre_exec` commands given;
        #   - runs the `pre_exec_cached` commands given;
        #   - dumps the resulting env in a temporary file;
        #
        # Then run that script and read the resulting env back into a dict to
        # return.  If `script_path` is specified, then also create a file at the
        # given name and fill it with `unset` and `export` statements to
        # recreate that specific environment: any shell sourcing that
        # `script_path` file thus activates the environment we just prepared.
        tmp_file, tmp_name = tempfile.mkstemp()

        # use a file object to simplify byte conversion
        fout = os.fdopen(tmp_file, 'w')
        try:
            fout.write('\n')
            if unset:
                fout.write('# unset\n')
                for k in sorted(unset):
                    if k not in environment:
                        fout.write('unset %s\n' % k)
                fout.write('\n')

            if BLACKLIST:
                fout.write('# blacklist\n')
                for k in sorted(BLACKLIST):
                    fout.write('unset %s\n' % k)
                fout.write('\n')

            if environment:
                fout.write('# export\n')
                for k in sorted(environment.keys()):
                    if k not in BLACKLIST:
                        fout.write("export %s=%s\n"
                                  % (k, _quote(environment[k])))
                fout.write('\n')

            if pre_exec_cached:
                fout.write('# pre_exec (cached)\n')
                # do not sort, order dependent
                for cmd in pre_exec_cached:
                    fout.write('%s\n' % cmd)
                fout.write('\n')

        finally:
            fout.close()

        cmd = '/bin/sh -c ". %s && /usr/bin/env | /usr/bin/sort"' % tmp_name
        out, err, ret = sh_callout(cmd)

        if ret:
            raise RuntimeError('error running "%s": %s' % (cmd, err))

        env = env_read_lines(out.split('\n'))
      # os.unlink(tmp_name)

        _env_cache[cache_md5] = env


    # If `script_path` is specified, create a script with that name which unsets
    # the same names as in the tmp script above, and exports all vars from the
    # resulting env from above (thus storing the *results* of the
    # `pre_exec_cached` env, not the env and `pre_exec_cached` directives
    # themselves).
    #
    # FIXME: files could also be cached and re-used (copied or linked)
    if script_path:
        with open(script_path, 'w') as fout:

            fout.write('\n# unset\n')
            for k in unset:
                if k not in sorted(environment):
                    fout.write('unset %s\n' % k)
            fout.write('\n')

            fout.write('# blacklist\n')
            for k in sorted(BLACKLIST):
                fout.write('unset %s\n' % k)
            fout.write('\n')

            fout.write('# export\n')
            for k in sorted(env.keys()):
                # FIXME: shell quoting for value
                fout.write("export %s=%s\n" % (k, _quote(env[k])))
            fout.write('\n')

            if pre_exec:
                # do not sort, order dependent
                fout.write('# pre_exec\n')
                for cmd in pre_exec:
                    fout.write('%s\n' % cmd)
                fout.write('\n')

    return env


# ------------------------------------------------------------------------------
#
def env_diff(env_1, env_2):
    '''
    This method serves debug purposes: it compares to environments and returns
    those elements which appear in only either one or the other env, and which
    changed from one env to another.
    '''

    only_1  = dict()
    only_2  = dict()
    changed = dict()

    keys_1 = sorted(env_1.keys())
    keys_2 = sorted(env_2.keys())

    for k in keys_1:
        v = env_1[k]
        if   k not in env_2: only_1[k]  = v
        elif v != env_2[k] : changed[k] = [v, env_2[k]]

    for k in keys_2:
        v = env_2[k]
        if k not in env_1: only_2[k]  = v
        # else is checked in keys_1 loop above

    return only_1, only_2, changed


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2018"
__license__   = "GPL"


from .server import Server
from .client import Client, Request,   Response
from .bridge import Bridge
from .queue  import Queue,  Putter,    Getter
from .pubsub import PubSub, Publisher, Subscriber


# ------------------------------------------------------------------------------





import zmq
import msgpack

import threading as mt

from ..json_io import read_json
from ..misc    import as_list

from .utils    import no_intr, sock_connect


# --------------------------------------------------------------------------
#
_LINGER_TIMEOUT    = 1000  # ms to linger after close
_HIGH_WATER_MARK   = 1024  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE =    1  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
class Request(object):

    def __init__(self, cmd=None, arg=None):

        self._cmd = cmd
        self._arg = arg


    @classmethod
    def from_dict(cls, req):

        return Request(cmd=req['cmd'], arg=req.get('arg'))


    def packb(self):

        msg_req = {'cmd': self._cmd, 'arg': self._arg}
        return msgpack.packb(msg_req)


    @property
    def cmd(self): return self._cmd

    @property
    def arg(self): return self._arg


# ------------------------------------------------------------------------------
#
class Response(object):

    # FIXME: inherit future
    def __init__(self, res=None, err=None, exc=None):

        self._res = res
        self._err = err
        self._exc = exc

    def __repr__(self):

        ret = ''
        if self._res: ret += 'res: %s  ' % self._res
        if self._err: ret += 'err: %s  ' % self._err
        if self._exc: ret += 'exc: %s  ' % self._exc[-1]

        return ret.strip()


    def str(self):

        if self._res: ret = 'res: %s' % self._res
        else        : ret = 'err: %s' % self._err

        return ret.strip()


    @classmethod
    def from_msg(cls, msg):
        return cls.from_dict(msgpack.unpackb(msg))


    @classmethod
    def from_dict(cls, msg):

        return Response(res=msg.get('res'),
                        err=msg.get('err'),
                        exc=msg.get('exc'))


    @property
    def res(self): return self._res

    @property
    def err(self): return self._err

    @property
    def exc(self): return as_list(self._exc)


# ------------------------------------------------------------------------------
#
class Client(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, server=None, url=None):

        if server:
            self._url = read_json('%s.cfg' % server)['addr']

        elif url:
            self._url = url

        else:
            raise ValueError('need server name/cfg or Url')

        self._cb   = None

        self._ctx  = zmq.Context()
        self._sock = self._ctx.socket(zmq.REQ)

        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK

        sock_connect(self._sock, self._url)

        self._term   = mt.Event()
        self._active = False


    @property
    def url(self):
        return self._url


    # --------------------------------------------------------------------------
    #
    def request(self, cmd=None, arg=None):

        req = Request(cmd=cmd, arg=arg)

        no_intr(self._sock.send, req.packb())

        res = Response.from_msg(no_intr(self._sock.recv))

        for l in res.exc:
            print(l)

        if res.err:
            raise RuntimeError('ERROR: %s' % res.err)

        return res


    # --------------------------------------------------------------------------
    #
    def close(self):

        self._sock.close()


# ------------------------------------------------------------------------------





import zmq
import msgpack

import threading as mt

from ..ids     import generate_id
from ..url     import Url
from ..misc    import as_string, get_hostip
from ..logger  import Logger
from ..profile import Profiler
from ..debug   import get_exception_trace

from .utils    import no_intr


# --------------------------------------------------------------------------
#
_LINGER_TIMEOUT    =         250  # ms to linger after close
_HIGH_WATER_MARK   = 1024 * 1024  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE =        1024  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
class Server(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, url=None):

        # this server offers only synchronous communication: a request will be
        # worked upon and answered before the next request is received.

        self._url    = url
        self._uid    = generate_id('server', ns='radical.utils')
        self._cbs    = dict()

        self._log    = Logger(self._uid, level='debug', targets='.')
        self._prof   = Profiler(self._uid, path='.')

        self._addr   = None
        self._thread = None
        self._up     = mt.Event()
        self._term   = mt.Event()

        self.register_request('echo', self._request_echo)
        self.register_request('fail', self._request_fail)

        if not self._url:
            self._url  = 'tcp://*:10000-11000'

        # URLs can specify port ranges to use - check if that is the case (see
        # default above) and initilize iterator.  The URL is expected to have
        # the form:
        #
        #   <proto>://<iface>:<ports>/
        #
        # where
        #   <proto>: any protocol accepted by zmq,       defaults to `tcp`
        #   <iface>: IP number of interface to bind to   defaults to `*`
        #   <ports>: port range to find port to bind to  defaults to `*`
        #
        # The port range can be formed as:
        #
        #   '*'      : any port
        #   '100+'   : any port equal or larger than 100
        #   '100-'   : any port equal or larger than 100
        #   '100-110': any port equal or larger than 100, up to 110
        tmp = self._url.split(':', 2)
        assert(len(tmp) == 3)
        self._proto = tmp[0]
        self._iface = tmp[1].lstrip('/')
        self._ports = tmp[2].replace('+', '-')

        tmp = self._ports.split('-')

        if len(tmp) == 0:
            self._port_start = 1
            self._port_stop  = None
        elif len(tmp) == 1:
            if tmp[0] == '*':
                self._port_this  = '*'
                self._port_start = None
                self._port_stop  = None
            else:
                self._port_start = int(tmp[0])
                self._port_stop  = int(tmp[0])
        elif len(tmp) == 2:
            if tmp[0]: self._port_start = int(tmp[0])
            else     : self._port_start = 1
            if tmp[1]: self._port_stop  = int(tmp[1])
            else     : self._port_stop  = None
        else:
            raise RuntimeError('cannot parse port spec %s' % self._ports)

        self._port_this = None


    # --------------------------------------------------------------------------
    #
    def _iterate_ports(self):

        if self._port_this == '*':
            # leave scanning to zmq
            yield self._port_this
            return


        if self._port_this is None:
            # initizliaze range iterator
            self._port_this = self._port_start

        if self._port_stop is None:
            while True:
                yield self._port_this
                self._port_this += 1

        else:
            while self._port_this <= self._port_start:
                yield self._port_this
                self._port_this += 1


    # --------------------------------------------------------------------------
    #
    def _iterate_urls(self):

        for port in self._iterate_ports():
            yield '%s://%s:%s' % (self._proto, self._iface, port)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    @property
    def addr(self):
        return self._addr


    # --------------------------------------------------------------------------
    #
    def start(self):

        self._log.info('start bridge %s', self._uid)

        if self._thread:
            raise RuntimeError('`start()` can be called only once')

        self._thread = mt.Thread(target=self._work)
        self._thread.start()

        self._up.wait()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._log.info('stop bridge %s', self._uid)
        self._term.set()


    # --------------------------------------------------------------------------
    #
    def wait(self):

        self._log.info('wait bridge %s', self._uid)

        if self._thread:
            self._thread.join()
        self._log.info('wait bridge %s', self._uid)


    # --------------------------------------------------------------------------
    #
    def register_request(self, req, cb):

        self._log.info('add handler: %s: %s', req, cb)
        self._cbs[req] = cb


    # --------------------------------------------------------------------------
    #
    def _request_fail(self, arg):

        raise RuntimeError('task failed successfully')


    # --------------------------------------------------------------------------
    #
    def _request_echo(self, arg):

        return {'res': arg}


    # --------------------------------------------------------------------------
    #
    def _success(self, res=None):

        return {'err': None,
                'exc': None,
                'res': res}


    # --------------------------------------------------------------------------
    #
    def _error(self, err=None, exc=None):

        if not err:
            err = 'invalid request'

        return {'err': err,
                'exc': exc,
                'res': None}


    # --------------------------------------------------------------------------
    #
    def _work(self):

        self._ctx  = zmq.Context()
        self._sock = self._ctx.socket(zmq.REP)

        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK

        for url in self._iterate_urls():
            try:
                self._log.debug('try url %s', url)
                self._sock.bind(url)
                self._log.debug('success')
                break
            except Exception:
                self._log.exception('pass')


        self._addr      = Url(as_string(self._sock.getsockopt(zmq.LAST_ENDPOINT)))
        self._addr.host = get_hostip()
        self._addr      = str(self._addr)

        self._up.set()

        self._poll = zmq.Poller()
        self._poll.register(self._sock, zmq.POLLIN)

        while not self._term.is_set():

            event = dict(no_intr(self._poll.poll, timeout=100))

            if self._sock not in event:
                continue

            data = no_intr(self._sock.recv)
            req  = msgpack.unpackb(data)
            self._log.debug('req: %s', req)

            if not isinstance(req, dict):
                rep = self._error(err='invalid message type')

            else:
                cmd = req.get('cmd')
                arg = req.get('arg')

                if not cmd:
                    rep = self._error(err='no command in request')

                elif cmd not in self._cbs:
                    rep = self._error(err='command unknown')

                elif not arg:
                    rep = self._error(err='missing arguments')

                else:
                    try:
                        rep = self._success(self._cbs[cmd](arg))
                    except Exception as e:
                        rep = self._error(err='command failed: %s' % str(e),
                                          exc=get_exception_trace())

            no_intr(self._sock.send, msgpack.packb(rep))
            self._log.debug('rep: %s', rep)

        self._log.debug('term')


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import os

import radical.utils as ru

# test environment specific
BLACKLIST_LOCAL_SYS = ['XPC_SERVICE_NAME']


# ------------------------------------------------------------------------------
#
def test_env_prep():

    try   : del(os.environ['TEST'])
    except: pass

    # clean `os.environ`
    for env_name in BLACKLIST_LOCAL_SYS:
        if env_name in os.environ:
            del os.environ[env_name]

    env = dict(os.environ)
    ret = ru.env_prep(env)
    _, only_ret, changed = ru.env_diff(env, ret)
    assert(not only_ret), only_ret
    assert(not changed), changed

    env = dict(os.environ)

    os.environ['OS_ONLY']  = 'os_only'
    os.environ['SHARED']   = 'os_shared'
    env       ['SHARED']   = 'env_shared'
    env       ['ENV_ONLY'] = 'env_only'

    ret = ru.env_prep(environment=env, unset=['OS_ONLY'], script_path='/tmp/test.env')
    assert('OS_ONLY' not in ret), ret['OS_ONLY']
    assert(ret['ENV_ONLY'] == 'env_only')
    assert(ret['SHARED'] == 'env_shared')

    only_env, only_ret, changed = ru.env_diff(env, ret)
    assert(not only_ret), [only_env, only_ret, changed]
    assert(not changed),  changed

    out, _, ret = ru.sh_callout('export OS_ONLY=x; . /tmp/test.env; echo $OS_ONLY',
                                 shell=True)
    out = out.strip()
    assert(not out), out
    assert(not ret), ret

    out, _, ret = ru.sh_callout('unset ENV_ONLY; . /tmp/test.env; echo $ENV_ONLY',
                                  shell=True)
    out = out.strip()
    assert(out == 'env_only'), out
    assert(not ret), ret


# ------------------------------------------------------------------------------
#
def test_env_read():

    fname = '/tmp/env.%d' % os.getpid()
    os.system('/bin/sh -c "env > %s"' % fname)
    try:
        env = ru.env_read(fname)

        for k,v in env.items():
            assert(os.environ[k] == v)

        for k,v in os.environ.items():
            if k not in ru.env.BLACKLIST:
                assert(env[k] == v)

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_env_eval():

    fname = '/tmp/env.%d' % os.getpid()
    try:
        cmd   = "sh -c \"env " \
                "| sed -e \\\"s/^/export /\\\" "\
                "| sed -e \\\"s/=/='/\\\" " \
                "| sed -e \\\"s/$/'/\\\" "\
                "> %s\"" % fname
        os.system(cmd)
        env = ru.env_eval(fname)

        for k,v in env.items():
            if k not in ru.env.BLACKLIST:
                assert(os.environ[k] == v), [os.environ[k], v]

        for k,v in os.environ.items():
            if k not in ru.env.BLACKLIST:
                assert(env[k] == v)

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_env_prep()
    test_env_read()
    test_env_eval()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import json
import time
import pprint

import radical.utils as ru


import pytest
yaml = pytest.importorskip('yaml')
flux = pytest.importorskip('flux')


# ------------------------------------------------------------------------------
#
def test_flux():

    def cb1(job_id, state, ts, context):
        print('=== event ===', end='')
        pprint.pprint([job_id, state, ts, context])

    fh   = None
    uid  = None
    info = None

    try:
        fh   = ru.FluxHelper()
        info = fh.start_service()
        uid  = info['uid']

        fh.register_callback(uid, cb1)

        print('\n--- start', uid)
        pprint.pprint(info)

        print('\n--- check', uid)
        pprint.pprint(fh.check_service(info['uid']))

        print('\n--- handle', uid)
        pprint.pprint(fh.get_handle(info['uid']))

        print('\n--- executor', uid)
        jex = fh.get_executor(info['uid'])
        pprint.pprint(jex)
        jobspec = json.dumps(ru.read_json('/tmp/spec.json'))
        fut   = jex.submit(jobspec, waitable=True)
        jobid = fut.jobid()
        print('\n--- job', uid, jobid)
        time.sleep(3)

        print('\n--- check', uid)
        pprint.pprint(fh.check_service(info['uid']))

        print('\n--- close', uid)
        pprint.pprint(fh.close_service(info['uid']))

    finally:

        if uid:
            try:
                print('\n--- final', uid)
                pprint.pprint(fh.check_service(info['uid']))
            except:
                print('\n--- ok')


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_flux()


# ------------------------------------------------------------------------------





# FIXME: this is work in progress and should not be used

from .proxy import Proxy  # pylint: disable=import-error


# ------------------------------------------------------------------------------
#
class PtyProxy(Proxy):

    '''

    Different RCT layers need to frequently interact with shell environments on
    remote resources.  This `PtyProxy` class provides that functionality.

    Thye class inherits from the `ru.Proxy` class, and accepts the same initial
    parameters which are then used to create a tunnel to the target resource.
    Once that tunnel is established, we'll bootstrap a small Python/ZMQ server
    endpoint on the target resource and connect to it.  That enspoint will be
    used to interface to the remote OS, by serving as a command execution
    service.

    The service endpoint is considered ephemeral, but it will keep some state on
    the target machine's disk space and thus can reconnect to that state.
    Commands which are executed by the service endpoint can survive the endpoint
    itself, and can also be revconnected to.  Any instance of the `PtyProcess`
    class will thus have access to the complete history of all previous
    instances targeting the same resource (the `purge()` method will clean out
    that history, and later incarnations may limit history by time or size).

    # FIXME: make this configurable

    The serive endpoint will create a Python virtualenv on the target resource
    on the fly (or reuse it when it exists).  The default location is in
    `$HOME/.radical/utils/ptyproxy/ve/`.

    # FIXME: make this configurable
    # FIXME: make resource details configurable (`module load python` etc.).
    #        Limit this to a bare minimum, ie. expect a POSIX shell and basic
    #        Python.

    The class also provides basic file transfer capablities.  Those should
    suffice to move scripts and coinfiguration files around, but are likely very
    inefficient for any larger and/or binary data transfer.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, url=None, timeout=None):

        # FIXME
      # super(Proxy, self).__init__(url=url, timeout=timeout)
        Proxy.__init__(url, timeout)

        self._connected = False


    # --------------------------------------------------------------------------
    #
    def is_connected(self, _raise=False):

        if _raise and not self._connected:
            raise RuntimeError('not connected')

        return self._connected


    # --------------------------------------------------------------------------
    #
    def purge(self):

        self.is_connected(_raise=True)

        ret = self._run_command({'cmd': 'purge'})
        if not ret.get('ret'):
            raise RuntimeError('purge failed: %s' % ret.get('error'))


    # --------------------------------------------------------------------------
    #
    def stage_to_remote(self, src, tgt=None):
        pass


    # --------------------------------------------------------------------------
    #
    def stage_from_remote(self, src, tgt=None):
        pass


# ------------------------------------------------------------------------------

